"""ACE. Option."""

from __future__ import annotations

import string
from functools import total_ordering

from cisco_acl import helpers as h
from cisco_acl.base import Base
from cisco_acl.types_ import LStr, DAny

LOGS = ("log", "log-input")


@total_ordering
class Option(Base):
    """ACE. Option."""

    def __init__(self, line: str = "", **kwargs):
        """Init Option.

        :param line: Option line.
        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        :type platform: str

        :param version: Software version, default is "0".
        :type version: str

        Helpers
        :param note: Object description.
        :type note: Any

        :example:
            option = Option("ack dscp ef log")
            option.line -> "ack log"
            option.flags -> {"ack", "dscp", "ef"}
            option.logs -> {"log"}
        """
        self._line: str = ""
        self._flags: LStr = []
        self._logs: LStr = []
        super().__init__(**kwargs)  # platform, note
        self.line = line

    # ========================== redefined ===========================

    def __hash__(self) -> int:
        """__hash__."""
        return tuple(self._flags).__hash__()

    def __eq__(self, other) -> bool:
        """== equality."""
        return self.__hash__() == other.__hash__()

    def __lt__(self, other) -> bool:
        """< less than."""
        if self.__class__ == other.__class__:
            return self._flags < self._flags
        return False

    # =========================== property ===========================

    @property
    def flags(self) -> LStr:
        """ACE Option flags, items that related to packet forwarding."""
        return self._flags

    @property
    def line(self) -> str:
        """ACE Option items as string."""
        return self._line

    @line.setter
    def line(self, line: str) -> None:
        line = h.init_line(line)
        self._line = line

        options: LStr = [s.strip() for s in line.split()]
        options = [s for s in options if s]
        for option in options:
            if option[0] not in string.ascii_lowercase:
                raise ValueError(f"invalid {option=}")

        self._flags = [s for s in options if s not in LOGS]
        self._logs = [s for s in options if s in LOGS]

    @property
    def logs(self) -> LStr:
        """ACE Option logs, items that not related to packet forwarding."""
        return self._logs

    # =========================== method =============================

    def data(self, uuid: bool = False) -> DAny:
        """Return Option data as dictionary.

        :param uuid: Return self.uuid in data.
        :type uuid: bool

        :return: Option data.

        :example:
        option = Option("ack log")
        option.data() -> {
            "line": "ack log",
            "platform": "ios",
            "version": "0",
            "note": "",
            "flags": ["ack"],
            "logs": ["log"],
        }
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            version=str(self.version),
            note=self.note,
            # property
            flags=self._flags,
            logs=self._logs,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data



"""AceBase, parent of: Ace, Remark, AceGroup."""

from abc import ABC, abstractmethod

from cisco_acl import helpers as h
from cisco_acl.base import Base
from cisco_acl.types_ import StrInt, DAny
from cisco_acl.wildcard import init_max_ncwb


class AceBase(Base, ABC):
    """AceBase, parent of: Ace, Remark, AceGroup."""

    def __init__(self, **kwargs):
        """Init AceBase.

        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        :type platform: str

        :param version: Software version, default is "0".
        :type version: str

        Helpers
        :param note: Object description.
        :type note: Any

        :param max_ncwb: Max count of non-contiguous wildcard bits.
        :type max_ncwb: int

        :param protocol_nr: Well-known ip protocols as numbers.
            True  - all ip protocols as numbers,
            False - well-known ip protocols as names (default).
        :type protocol_nr: bool

        :param port_nr: Well-known TCP/UDP ports as numbers.
            True  - all tcp/udp ports as numbers,
            False - well-known tcp/udp ports as names (default).
        :type port_nr: bool

        Alternate way to get `name` and ACEs `items`, if `line` absent.
        :param str type: ACL type: "extended", "standard" (default "extended").
        """
        self._line: str = ""
        self._sequence: int = 0
        self._type: str = "extended"
        self._protocol_nr: bool = False
        self._port_nr: bool = False
        # noinspection PyProtectedMember
        self.max_ncwb: int = init_max_ncwb(**kwargs)
        super().__init__(**kwargs)  # platform, note
        if kwargs.get("type"):
            self._type = h.init_type(**kwargs)
        if sequence := kwargs.get("sequence"):
            self._sequence = h.init_int(sequence)
        if protocol_nr := kwargs.get("protocol_nr"):
            self._protocol_nr = bool(protocol_nr)
        if port_nr := kwargs.get("port_nr"):
            self._port_nr = bool(port_nr)

    def __hash__(self) -> int:
        """__hash__."""
        return self.line.__hash__()

    def __eq__(self, other) -> bool:
        """== equality."""
        if self.__class__ == other.__class__:
            return self.__hash__() == other.__hash__()
        return False

    def __repr__(self):
        """__repr__."""
        name = self.__class__.__name__
        params = self._repr__params()
        params = self._repr__add_param("protocol_nr", params)
        params = self._repr__add_param("port_nr", params)
        kwargs = ", ".join(params)
        return f"{name}({kwargs})"

    # =========================== property ===========================

    @property
    def port_nr(self) -> bool:
        """Well-known TCP/UDP ports as numbers."""
        return self._port_nr

    @port_nr.setter
    def port_nr(self, port_nr: bool) -> None:
        self._port_nr = bool(port_nr)
        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    @property
    def protocol_nr(self) -> bool:
        """Well-known ip protocols as numbers."""
        return self._protocol_nr

    @protocol_nr.setter
    def protocol_nr(self, protocol_nr: bool) -> None:
        self._protocol_nr = bool(protocol_nr)
        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    @property
    def sequence(self) -> int:
        """ACE sequence number in ACL.

        :return: Sequence number.

        :example: Ace without sequence number
            self: Ace("permit ip any any")
            return: 0

        :example: Ace with sequence number.
            ace = Ace("10 permit ip any any")
            ace.sequence -> 10
        """
        return self._sequence

    @sequence.setter
    def sequence(self, sequence: StrInt) -> None:
        self._sequence = h.init_int(sequence)

    @property
    def type(self) -> str:
        """ACL type: standard, extended."""
        return self._type

    @type.setter
    def type(self, type_: str) -> None:
        self._type = h.init_type(type=type_, platform=self.platform)
        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    # =========================== method =============================

    def copy(self):
        """Copy the self object."""
        kwargs = self.data()
        return self.__class__(**kwargs)

    @abstractmethod
    def data(self, uuid: bool = False) -> DAny:
        """Convert self object to the dictionary.

        :param uuid: Return self.uuid in data.
        :type uuid: bool

        :return: The dictionary.
        """

    # =========================== helper =============================

    def _sequence_s(self) -> str:
        """Return string of sequence, empty string if sequence==0."""
        if self._sequence:
            return str(self._sequence)
        return ""



"""Group of ACE (Access Control Entry).

These are multiple ACEe items, which must be in a certain order.
If you are changing Ace items order (sequence numbers) inside Acl,
the AceGroup behaves like a single item and order of ACE items inside AceGroup is not changed.
AceGroup is useful for freezing ACEs section, to hold "deny" after certain "permit".
This class implements most of the Python list methods: append(), extend(), sort(), etc.
"""

from __future__ import annotations

import logging
from functools import total_ordering
from ipaddress import NetmaskValueError
from typing import List, Optional, Union

from cisco_acl import parsers, helpers as h
from cisco_acl.ace import Ace, LAce
from cisco_acl.ace_base import AceBase
from cisco_acl.group import Group
from cisco_acl.helpers import ACTIONS
from cisco_acl.remark import Remark, LRemark
from cisco_acl.types_ import DAny
from cisco_acl.wildcard import init_max_ncwb

UAce = Union[Ace, Remark]
USAce = Union[Ace, Remark, str]

OUAce = Optional[UAce]
LUAce = List[UAce]
LUSAce = List[USAce]

ULAce = Union[LAce, LRemark, LUAce]


@total_ordering
class AceGroup(AceBase, Group):
    """Group of ACE (Access Control Entry)."""

    def __init__(self, line: str = "", **kwargs):
        r"""Init AceGroup.

        This class implements most of the Python list methods: append(), extend(), sort(), etc.
        :param line: String of ACEs, lines that starts with "allow", "deny", "remark".
        :type line: str

        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        :type platform: str

        :param version: Software version, default is "0".
        :type version: str

        Helpers
        :param note: Object description.
        :type note: Any

        :param max_ncwb: Max count of non-contiguous wildcard bits.
        :type max_ncwb: int

        :param protocol_nr: Well-known ip protocols as numbers.
            True  - all ip protocols as numbers,
            False - well-known ip protocols as names (default).
        :type protocol_nr: bool

        :param port_nr: Well-known TCP/UDP ports as numbers.
            True  - all tcp/udp ports as numbers,
            False - well-known tcp/udp ports as names (default).
        :type port_nr: bool

        :param group_by: Startswith in remark line. ACEs group, starting from the Remark,
            where line startswith `group_by`, will be applied to the same AceGroup,
            until next Remark that also startswith `group_by`.
        :type group_by: str

        Alternate way to get `name` and ACEs `items`, if `line` absent.
        :param str type: ACL type: "extended", "standard" (default "extended").

        :param name: Name of AceGroup, usually Remark.text of 1st self.items.
        :type name: str

        :param items: An alternate way to create AceGroup object from a list of Ace.
            (default from `line`)
        :type items: List[Union[Ace, Remark, str]]

        :example:
            aceg = AceGroup("10 permit icmp any any\n  20 deny ip any any")
            aceg.line -> "10 permit icmp any any\n20 deny ip any any"
            aceg.platform -> "ios"
            aceg.note -> "description"
            aceg.sequence -> 20  # Taking from the first ACE in items.
            aceg.items -> [Ace("10 permit icmp any any"), Ace("20 deny ip any any")]
        """
        self._name = ""
        self._group_by = ""
        self._items: LUAce = []
        AceBase.__init__(self, **kwargs)  # platform, type, note, sequence, protocol_nr, port_nr
        Group.__init__(self)
        if name := kwargs.get("name"):
            self._name = h.init_name(name)
        if group_by := str(kwargs.get("group_by") or ""):
            self._group_by = group_by
        # noinspection PyProtectedMember
        self.max_ncwb: int = init_max_ncwb(**kwargs)
        if items := kwargs.get("items") or []:
            self.items = items
            return
        self.line = line

    def __lt__(self, other) -> bool:
        """< less than."""
        if hasattr(other, "sequence"):
            if self._sequence == other.sequence:
                if other.__class__.__name__ == "Remark":
                    return False
                if isinstance(other, Ace):
                    return False
                if isinstance(other, AceGroup):
                    return str(self) < str(other)
                raise TypeError(f"{other=} {AceGroup} expected")
            return self._sequence < other.sequence
        return False

    # =========================== property ===========================

    @property
    def group_by(self) -> str:
        """Group ACEs to AceGroup by startswith ot this value in remarks."""
        return self._group_by

    @property
    def items(self) -> LUAce:
        """List of ACE items: Ace, Remark, AceGroup."""
        return self._items

    @items.setter
    def items(self, items: LUSAce) -> None:
        if isinstance(items, (str, Ace, Remark)):
            items = [items]
        if not isinstance(items, (list, tuple)):
            raise TypeError(f"{items=} {list} expected")

        _items: LUAce = []  # result
        for item in items:
            if isinstance(item, (Ace, Remark)):
                item._platform = self._platform
                item.version = self.version
                item._type = self._type
                _items.append(item)
            elif isinstance(item, dict):
                item["platform"] = self._platform
                item["version"] = str(self.version)
                item["type"] = self._type
                ace_o: UAce = self._dict_to_ace(**item)
                _items.append(ace_o)
            elif isinstance(item, str):
                line = h.init_line(item)
                item_ = self._line_to_ace(line)
                _items.append(item_)
            else:
                raise TypeError(f"{item=} {str} expected")
        self._items = _items

    @property
    def line(self) -> str:
        """Group of ACL config line."""
        lines = [o.line for o in self._items]
        return "\n".join(lines)

    @line.setter
    def line(self, line: str) -> None:
        lines = line.split("\n")
        lines = [h.init_line(s) for s in lines]
        items_ = [self._line_to_oace(s, warning=True) for s in lines]
        items: LUAce = [o for o in items_ if isinstance(o, (Ace, Remark))]
        self.items = items
        sequence = self._items[0].sequence if self._items else 0
        self.sequence = sequence

    @property
    def name(self) -> str:
        """Acl/AceGroup name."""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """Acl/AceGroup name, without "ip access-list ".

        Requirements:
        - length <= 100 chars.
        - all chars are digits.
        - first char is ascii_letters, other chars are ascii_letters and punctuation.
        """
        self._name = h.init_name(name)

    @property
    def platform(self) -> str:
        """Platform.

        - "ios" Cisco IOS (extended ACL).
        - "nxos" Cisco Nexus NX-OS.
        """
        return self._platform

    @platform.setter
    def platform(self, platform: str) -> None:
        """Change platform, normalizes self.items regarding the new platform.

        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        """
        self._platform = h.init_platform(platform=platform)

        for item in self._items:
            item.type = self._type
            if self._platform == "nxos":
                self.ungroup_ports()
            item.platform = self._platform

        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    @property
    def type(self) -> str:
        """ACL type: standard, extended."""
        return self._type

    @type.setter
    def type(self, type_: str) -> None:
        type_ = h.init_type(type=type_, platform=self.platform)
        for item in self._items:
            item.type = type_
        self._type = type_

        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    # =========================== method =============================

    def data(self, uuid: bool = False) -> DAny:
        """Convert AceGroup object to the dictionary.

        :param uuid: Return self.uuid in data.
        :type uuid: bool

        :return: The dictionary.
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            version=str(self.version),
            type=self._type,
            name=self._name,
            items=[o.data(uuid=uuid) for o in self._items],
            group_by=self._group_by,
            note=self.note,
            protocol_nr=self._protocol_nr,
            port_nr=self._port_nr,
            # property
            sequence=self._sequence,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data

    def delete_note(self) -> None:
        """Delete note in all children self.items: Ace, AceGroup, Remark."""
        self.note = ""
        for item in self._items:
            item.note = ""
            if isinstance(item, AceGroup):
                item.delete_note()

    def get_remark_name(self) -> str:
        """Return Rule name from 1st remark."""
        remark = self.get_remark()
        name = h.parse_remark_name(text=remark.text, group_by=self.group_by)
        return name

    def get_remark(self) -> Remark:
        """Return 1st remark with Rule name."""
        remark = self.items[0]
        if not isinstance(remark, Remark):
            raise TypeError(f"{remark=} {Remark} expected")
        return remark

    def tcam_count(self) -> int:
        """Calculate sum of ACEs. Also takes into account the addresses in the address group.

        Useful for getting an estimate of the amount of TCAM resources needed for this ACL.
        :return: Count of TCAM resources.
        """
        counter = 0
        for item in self.items:
            if isinstance(item, AceGroup):
                counter += item.tcam_count()
                continue
            if not isinstance(item, Ace):
                continue
            if "addrgroup" not in [item.srcaddr.type, item.dstaddr.type]:
                counter += 1
                continue
            src_counter = 1
            if item.srcaddr.type == "addrgroup":
                src_counter = len(item.srcaddr.items) or 1
            dst_counter = 1
            if item.dstaddr.type == "addrgroup":
                dst_counter = len(item.dstaddr.items) or 1
            counter += src_counter * dst_counter
        return counter

    # noinspection PyIncorrectDocstring
    @h.check_start_step_sequence
    def resequence(self, start: int = 10, step: int = 10, **kwargs) -> int:
        """Resequence all AceGroup.items and change sequence numbers.

        :param start: Starting sequence number. start=0 - delete all sequence numbers.
        :param step: Step to increment the sequence number.
        :param items: List of Ace objects (default self.items).
        :return: Last sequence number.
        """
        items: LUAce = kwargs.get("items") or self._items
        sequence: int = int(start)
        count = len(items)

        for id_, item in enumerate(items, start=1):
            if isinstance(item, AceGroup):
                params = dict(items=item.items)
                sequence = self.resequence(start=sequence, step=step, **params)
            item.sequence = sequence
            if id_ < count:
                sequence += step
        return sequence

    def ungroup_ports(self) -> None:
        """Ungroup ports.

        ACEs with multiple ports in single line ("eq" or "neq") to multiple lines with single port.
        :example:
            aceg = AceGroup("permit tcp any eq 1 2 any eq 3 4")
            aceg.split_ports()
            aceg.line -> "permit tcp any eq 1 any eq 3
                          permit tcp any eq 1 any eq 4
                          permit tcp any eq 2 any eq 3
                          permit tcp any eq 2 any eq 4"
        """
        _items: LUAce = []
        for ace_o in self._items:
            if isinstance(ace_o, Ace):
                aces: LAce = ace_o.ungroup_ports()
                _items.extend(aces)
                continue
            _items.append(ace_o)
        self.items = _items

    # =========================== helper =============================

    def _dict_to_ace(self, **kwargs) -> UAce:
        """Convert dict data to object: Ace, Remark.

        :param kwargs: ACE data.
        :return: ACE object: Ace or Remark.

        :example: permit.
            data: {line="permit ip any any"}
            return: Ace("permit ip any any")

        :example: remark.
            data: {line="remark text"}
            return: Remark("text")
        """
        action = kwargs.get("action")
        if action not in ACTIONS:
            raise ValueError(f"invalid action in {kwargs=}, expected {ACTIONS=}")
        kwargs["platform"] = self._platform
        kwargs["version"] = str(self.version)
        kwargs["type"] = self._type
        kwargs["protocol_nr"] = self._protocol_nr
        kwargs["port_nr"] = self._port_nr
        if action == "remark":
            return Remark(**kwargs)
        return Ace(**kwargs)

    def _dict_to_aceg(self, **kwargs) -> UAceg:
        """Convert dict data to object: AceGroup, Ace, Remark.

        :param kwargs: ACE data.
        :return: ACE object: AceGroup, Ace, Remark.
        """
        items = kwargs.get("items")
        if isinstance(items, (list, tuple)):
            kwargs["platform"] = self._platform
            kwargs["version"] = str(self.version)
            kwargs["type"] = self._type
            kwargs["protocol_nr"] = self._protocol_nr
            kwargs["port_nr"] = self._port_nr
            return AceGroup(**kwargs)
        return self._dict_to_ace(**kwargs)

    def _line_to_ace(self, line: str) -> UAce:
        """Convert config line to object: Ace, Remark.

        :param line: ACE line.
        :return: ACE object: Ace or Remark.

        :example: permit
            line: "permit ip any any"
            return: Ace("permit ip any any")

        :example: remark
            line: "remark text"
            return: Remark("text")

        :example: not ACE line
            line: "text"
            return: None
        """
        action = parsers.parse_action(line)["action"]
        if action == "remark":
            return Remark(line, platform=self._platform, version=self.version, type=self._type)
        ace_o = Ace(
            line=line,
            platform=self._platform,
            version=self.version,
            type=self._type,
            protocol_nr=self._protocol_nr,
            port_nr=self._port_nr,
            max_ncwb=self.max_ncwb,
        )
        return ace_o

    def _line_to_oace(self, line: str, warning: bool = False) -> OUAce:
        """Convert config line to object: Ace, Remark, None."""
        if not line:
            return None

        if h.is_line_for_acl(line):
            try:
                ace_o: UAce = self._line_to_ace(line)
            except NetmaskValueError:
                raise
            except ValueError as ex:
                if warning:
                    msg = f"{type(ex).__name__}: {ex}. {line=} does not match ACE pattern"
                    logging.warning(msg)
                return None
            return ace_o

        known_skip = ["statistics ", "description ", "ignore "]
        for skip in known_skip:
            if line.startswith(skip):
                return None

        if warning:
            msg = f"{line=} does not match ACE pattern"
            logging.warning(msg)
        return None


LAceGroup = List[AceGroup]
UAceg = Union[Ace, Remark, AceGroup]
LUAceg = List[UAceg]



"""TCP/UDP ports and names mapping for Cisco ACL."""

from netports import SwVersion

from cisco_acl import helpers as h
from cisco_acl.types_ import DInt, DiStr, LStr, UVersion

# cisco Nexus 3172T, NXOS: version 9.3(8)
# cisco ISR4331/K9, Cisco IOS XE Software, Version 16.09.06
TCP_NAME_PORT__BASE = {
    "echo": 7,
    "discard": 9,
    "daytime": 13,
    "chargen": 19,
    "ftp-data": 20,
    "ftp": 21,
    "telnet": 23,
    "smtp": 25,
    "time": 37,
    "whois": 43,
    "tacacs": 49,
    "domain": 53,
    "gopher": 70,
    "finger": 79,
    "www": 80,
    "hostname": 101,
    "pop2": 109,
    "pop3": 110,
    "sunrpc": 111,
    "ident": 113,
    "nntp": 119,
    "bgp": 179,
    "irc": 194,
    "pim-auto-rp": 496,
    "exec": 512,
    "login": 513,
    "cmd": 514,
    "lpd": 515,
    "talk": 517,
    "uucp": 540,
    "klogin": 543,
    "kshell": 544,
}

# cisco ASA-5585 9.12(4)37
TCP_NAME_PORT__ASA = {
    "echo": 7,
    "discard": 9,
    "daytime": 13,
    "chargen": 19,
    "ftp-data": 20,
    "ftp": 21,
    "ssh": 22,
    "telnet": 23,
    "smtp": 25,
    "whois": 43,
    "tacacs": 49,
    "domain": 53,
    "gopher": 70,
    "finger": 79,
    "www": 80,
    "hostname": 101,
    "pop2": 109,
    "pop3": 110,
    "sunrpc": 111,
    "ident": 113,
    "nntp": 119,
    "netbios-ssn": 139,
    "imap4": 143,
    "bgp": 179,
    "irc": 194,
    "ldap": 389,
    "https": 443,
    "pim-auto-rp": 496,
    "exec": 512,
    "login": 513,
    "rsh": 514,
    "lpd": 515,
    "talk": 517,
    "uucp": 540,
    "klogin": 543,
    "kshell": 544,
    "rtsp": 554,
    "ldaps": 636,
    "kerberos": 750,
    "lotusnotes": 1352,
    "citrix-ica": 1494,
    "sqlnet": 1521,
    "h323": 1720,
    "pptp": 1723,
    "nfs": 2049,
    "ctiqbe": 2748,
    "cifs": 3020,
    "sip": 5060,
    "aol": 5190,
    "pcanywhere-data": 5631,
}

# cisco Nexus 3172T, NXOS: version 9.3(8)
TCP_NAME_PORT__NXOS = {
    "drip": 3949,
}
TCP_NAME_PORT__NXOS = {**TCP_NAME_PORT__BASE, **TCP_NAME_PORT__NXOS}

# cisco C6816-X-LE, Cisco IOS Software, Version 15.2(02)SY
TCP_NAME_PORT__IOS_15 = {
    "syslog": 514,  # alias of cmd
}
TCP_NAME_PORT__IOS_15 = {**TCP_NAME_PORT__BASE, **TCP_NAME_PORT__IOS_15}

# cisco ISR4331/K9, Cisco IOS XE Software, Version 16.09.06
TCP_NAME_PORT__IOS_16 = {
    "msrpc": 135,
    "syslog": 514,  # alias of cmd
    "onep-plain": 15001,
    "onep-tls": 15002,
}
TCP_NAME_PORT__IOS_16 = {**TCP_NAME_PORT__BASE, **TCP_NAME_PORT__IOS_16}

# cisco Nexus 3172T, NXOS: version 9.3(8)
# cisco ISR4331/K9, Cisco IOS XE Software, Version 16.09.06
UDP_NAME_PORT__BASE = {
    "echo": 7,
    "discard": 9,
    "time": 37,
    "nameserver": 42,
    "tacacs": 49,
    "domain": 53,
    "bootps": 67,
    "bootpc": 68,
    "tftp": 69,
    "sunrpc": 111,
    "ntp": 123,
    "netbios-ns": 137,
    "netbios-dgm": 138,
    "netbios-ss": 139,
    "snmp": 161,
    "snmptrap": 162,
    "xdmcp": 177,
    "dnsix": 195,
    "mobile-ip": 434,
    "pim-auto-rp": 496,
    "isakmp": 500,
    "biff": 512,
    "who": 513,
    "syslog": 514,
    "talk": 517,
    "rip": 520,
    "non500-isakmp": 4500,
}
# ASA
UDP_NAME_PORT__ASA = {
    "echo": 7,
    "discard": 9,
    "time": 37,
    "nameserver": 42,
    "tacacs": 49,
    "domain": 53,
    "bootps": 67,
    "bootpc": 68,
    "tftp": 69,
    "www": 80,
    "sunrpc": 111,
    "ntp": 123,
    "netbios-ns": 137,
    "netbios-dgm": 138,
    "snmp": 161,
    "snmptrap": 162,
    "xdmcp": 177,
    "dnsix": 195,
    "mobile-ip": 434,
    "pim-auto-rp": 496,
    "isakmp": 500,
    "biff": 512,
    "who": 513,
    "syslog": 514,
    "talk": 517,
    "rip": 520,
    "kerberos": 750,
    "radius": 1645,
    "radius-acct": 1646,
    "nfs": 2049,
    "cifs": 3020,
    "vxlan": 4789,
    "sip": 5060,
    "secureid-udp": 5510,
    "pcanywhere-status": 5632,
}
# cisco Nexus 3172T, NXOS: version 9.3(8)
UDP_NAME_PORT__NXOS = UDP_NAME_PORT__BASE

# cisco C6816-X-LE, Cisco IOS Software, Version 15.2(02)SY
UDP_NAME_PORT__IOS_15 = {**UDP_NAME_PORT__BASE}

# cisco ISR4331/K9, Cisco IOS XE Software, Version 16.09.06
UDP_NAME_PORT__IOS_16 = {
    "ripv6": 521,
}
UDP_NAME_PORT__IOS_16 = {**UDP_NAME_PORT__BASE, **UDP_NAME_PORT__IOS_16}


class PortName:
    """TCP/UDP ports and names mapping for Cisco ACL."""

    def __init__(
        self,
        protocol: str = "tcp",
        platform: str = "",
        version: UVersion = "",
    ):
        """Init PortName.

        :param protocol: Protocol: "tcp", "udp".
        :type protocol: str

        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        :type platform: str

        :param version: Software version.
        :type version: str
        """
        self.protocol: str = _init_protocol(protocol)
        self.platform: str = h.init_platform(platform=platform)
        self.version: SwVersion = h.init_version(version=version)

    def __repr__(self):
        """__repr__."""
        name = self.__class__.__name__
        params = [
            f"protocol={self.protocol!r}",
            f"platform={self.platform!r}",
        ]
        if str(self.version) != "0":
            params.append(f"version={str(self.version)!r}")
        params_s = ", ".join(params)
        return f"{name}({params_s})"

    # noinspection DuplicatedCode
    def names(self) -> DInt:
        """Return TCP/UDP protocol names and ports based on the platform and software version.

        :return: Dictionary with protocol names and ports (platform-specific).
        :example: {"echo": 7, "discard": 9, ...}
        """
        names_d: DInt = {}
        if self._is_tcp():
            if self.platform == "asa":
                names_d = TCP_NAME_PORT__ASA.copy()
            elif self.platform == "ios":
                if self.version.major == 15:
                    names_d = TCP_NAME_PORT__IOS_15.copy()
                else:
                    names_d = TCP_NAME_PORT__IOS_16.copy()
            elif self.platform == "nxos":
                names_d = TCP_NAME_PORT__NXOS.copy()

        elif self._is_udp():
            if self.platform == "asa":
                names_d = UDP_NAME_PORT__ASA.copy()
            elif self.platform == "ios":
                if self.version.major == 15:
                    names_d = UDP_NAME_PORT__IOS_15.copy()
                else:
                    names_d = UDP_NAME_PORT__IOS_16.copy()
            elif self.platform == "nxos":
                names_d = UDP_NAME_PORT__NXOS.copy()

        return names_d

    # noinspection DuplicatedCode
    def ports(self) -> DiStr:
        """Return TCP/UDP protocol ports and names based on the platform and software version.

        :return: Dictionary with protocol ports and ports (platform-specific).
        :example: {7: "echo", 9: "discard", ...}
        """
        names_d: DInt = self.names()
        ports_d: DiStr = _swap(names_d)
        return ports_d

    def _is_tcp(self) -> bool:
        """Check if the protocol is TCP.

        :return: True if the protocol is TCP, False otherwise.
        """
        return bool(self.protocol in ["tcp", "6"])

    def _is_udp(self) -> bool:
        """Check if the protocol is UDP.

        :return: True if the protocol is UDP, False otherwise.
        """
        return bool(self.protocol in ["udp", "17"])


# ============================ functions =============================


def all_known_names() -> LStr:
    """Return all known names, that can be used in Cisco ACL (platform does not matter)."""
    items = set()
    # tcp
    items.update(set(TCP_NAME_PORT__BASE))
    items.update(set(TCP_NAME_PORT__ASA))
    items.update(set(TCP_NAME_PORT__IOS_16))
    items.update(set(TCP_NAME_PORT__NXOS))
    items.update(set(UDP_NAME_PORT__BASE))
    # udp
    items.update(set(UDP_NAME_PORT__BASE))
    items.update(set(UDP_NAME_PORT__ASA))
    items.update(set(UDP_NAME_PORT__IOS_16))
    items.update(set(UDP_NAME_PORT__NXOS))
    return sorted(items)


# ============================= helpers ==============================


def _init_protocol(protocol: str) -> str:
    """Init protocol.

    Convert TCP/UDP numbers to names.
    :param protocol: Protocol: "tcp", "udp".
    :return: Protocol: "tcp", "udp".
    """
    protocol = str(protocol).lower()
    expected = ["tcp", "udp"]
    if protocol in expected:
        return protocol
    if str(protocol) == "6":
        return "tcp"
    if str(protocol) == "17":
        return "udp"
    raise ValueError(f"invalid {protocol=}, {expected=}")


def _swap(name_port: DInt) -> DiStr:
    """Swap names and ports in dict."""
    data: DiStr = {}
    for name, port in name_port.items():
        if port not in data:
            data[port] = name
    return data



"""ACE IP protocol object."""

from __future__ import annotations

from functools import total_ordering
from typing import List

from cisco_acl import helpers as h
from cisco_acl.base import Base
from cisco_acl.types_ import StrInt, DAny

PROTOCOL_IP = 0

OPTIONS = (
    "ack",
    "dscp",
    "fin",
    "log",
    "log-input",
    "match-all",
    "match-any",
    "precedence",
    "psh",
    "rst",
    "time-range",
    "tos",
    "ttl",
    "urg",
)
PROTOCOLS_IOS = {
    "ip": 0,
    "icmp": 1,
    "igmp": 2,
    "ipip": 4,
    "tcp": 6,
    "egp": 8,
    "udp": 17,
    "ipv6": 41,
    "gre": 47,
    "esp": 50,
    "ah": 51,
    "ahp": 51,
    "eigrp": 88,
    "ospf": 89,
    "nos": 94,
    "pim": 103,
    "pcp": 108,
}
PROTOCOLS_NXOS = {
    "ip": 0,
    "icmp": 1,
    "igmp": 2,
    "tcp": 6,
    "udp": 17,
    "gre": 47,
    "esp": 50,
    "ahp": 51,
    "eigrp": 88,
    "ospf": 89,
    "nos": 94,
    "pim": 103,
    "pcp": 108,
}
PROTOCOLS_ASA = {
    "ip": 0,
    "icmp": 1,
    "igmp": 2,
    "ipinip": 4,
    "tcp": 6,
    "igrp": 9,
    "udp": 17,
    "gre": 47,
    "esp": 50,
    "ah": 51,
    "icmp6": 58,
    "eigrp": 88,
    "ospf": 89,
    "nos": 94,
    "pim": 103,
    "pcp": 108,
    "snp": 109,
    "sctp": 132,
}
PROTOCOLS_ANY = {**PROTOCOLS_ASA, **PROTOCOLS_IOS, **PROTOCOLS_NXOS}
PROTOCOL_TO_NR = dict(
    asa=PROTOCOLS_ASA,
    ios=PROTOCOLS_IOS,
    nxos=PROTOCOLS_NXOS,
)
NR_TO_PROTOCOL = dict(
    asa={i: s for s, i in PROTOCOLS_ASA.items()},
    ios={i: s for s, i in PROTOCOLS_IOS.items()},
    nxos={i: s for s, i in PROTOCOLS_NXOS.items()},
)


@total_ordering
class Protocol(Base):
    """ACE IP protocol object."""

    def __init__(self, line: str = "", **kwargs):
        """Init Protocol.

        :param line: IP protocol line.
        :type line: str

        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        :type platform: str

        :param version: Software version, default is "0".
        :type version: str

        Helpers
        :param note: Object description.
        :type note: Any

        :param protocol_nr: Well-known ip protocols as numbers.
            True  - all ip protocols as numbers,
            False - well-known ip protocols as names (default).
        :type protocol_nr: bool

        :param has_port: ACE has tcp/udp src/dst ports.
            True  - ACL has tcp/udp src/dst ports,
            False - ACL does not have tcp/udp src/dst ports (default).
        :type port_nr: bool

        :example:
            protocol = Protocol("tcp", platform="ios")
            protocol.line -> "tcp"
            protocol.name -> "tcp"
            protocol.number -> 6

            protocol = Protocol("255", platform="ios")
            protocol.line -> "255"
            protocol.name == ""
            protocol.number == 255
        """
        self._number = PROTOCOL_IP
        self._protocol_nr = bool(kwargs.get("protocol_nr"))
        self._has_port = bool(kwargs.get("has_port"))
        super().__init__(**kwargs)
        self.line = line

    # ========================== redefined ===========================

    def __hash__(self) -> int:
        """__hash__."""
        return self.line.__hash__()

    def __eq__(self, other) -> bool:
        """== equality."""
        return self.__hash__() == other.__hash__()

    def __lt__(self, other) -> bool:
        """< less than."""
        if self.__class__ == other.__class__:
            if self.number != other.number:
                return self.number < other.number
        return False

    def __repr__(self):
        """__repr__."""
        params = super()._repr__params()
        params = self._repr__add_param("protocol_nr", params)
        params = self._repr__add_param("has_port", params)
        kwargs = ", ".join(params)
        name = self.__class__.__name__
        return f"{name}({kwargs})"

    # =========================== property ===========================

    @property
    def has_port(self) -> bool:
        """Return True if protocol is "tcp" or "udp" with specified port."""
        return self._has_port

    @has_port.setter
    def has_port(self, has_port: bool) -> None:
        self._has_port = bool(has_port)

    @property
    def line(self) -> str:
        """ACE protocol name: "ip", "icmp", "tcp", etc."""
        number = self._number
        if self._protocol_nr and not self._has_port:
            return str(number)
        if name := NR_TO_PROTOCOL[self._platform].get(number):
            return str(name)
        return str(number)

    @line.setter
    def line(self, line: str) -> None:
        line = h.int_to_str(line)
        if not line:
            self._number = PROTOCOL_IP
            return

        # permit 255 any any
        if line.isdigit():
            number: int = int(line)
            if not 0 <= number <= 255:
                raise ValueError(f"invalid protocol {number=}, expected 0..255")

        # permit ip any any
        else:
            number_ = PROTOCOLS_ANY.get(line)
            if number_ is None:
                raise ValueError(f"invalid protocol {line=}, expected={list(PROTOCOLS_ANY)}")
            number = int(number_)

        self._number = number

    @property
    def name(self) -> str:
        """ACE protocol name: "ip", "icmp", "tcp", etc."""
        number = self._number
        if name := NR_TO_PROTOCOL[self._platform].get(number):
            return str(name)
        return ""

    @name.setter
    def name(self, name: str) -> None:
        if not isinstance(name, str):
            raise TypeError(f"{name=} {str} expected")
        self.line = name

    @property
    def number(self) -> int:
        """ACE protocol number: 0..255, where 0="ip", 1="icmp", etc."""
        return self._number

    @number.setter
    def number(self, number: StrInt) -> None:
        self.line = h.init_number(number)

    @property
    def protocol_nr(self) -> bool:
        """Well-known ip protocols as numbers.

        True  - all ip protocols as numbers,
        False - well-known ip protocols as names (default).
        """
        return self._protocol_nr

    @protocol_nr.setter
    def protocol_nr(self, protocol_nr: bool) -> None:
        self._protocol_nr = bool(protocol_nr)

    # =========================== method =============================

    def data(self, uuid: bool = False) -> DAny:
        """Convert Protocol object to the dictionary.

        :param uuid: Return self.uuid in data.
        :return: Protocol data.

        :example:
            address = Protocol(line="tcp")
            address.data() -> {
                "line": "tcp",
                "platform": "ios",
                "version": "0",
                "note": "",
                "protocol_nr": False,
                "has_port": False,
                "name": "tcp",
                "number": 6,
            }
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            version=str(self.version),
            note=self.note,
            protocol_nr=self._protocol_nr,
            has_port=self._has_port,
            # property
            name=self.name,
            number=self._number,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data


LProtocol = List[Protocol]



"""Base - Parent of: Address, Port, Protocol, Ace, AceBase, Acl, AceGroup."""

from abc import ABC, abstractmethod
from typing import Any
from uuid import uuid1

from netports import SwVersion

from cisco_acl import helpers as h
from cisco_acl.helpers import IOS
from cisco_acl.types_ import LStr, DAny


class Base(ABC):
    """Base - Parent of: Address, Port, Protocol, Ace, AceBase, Acl, AceGroup."""

    def __init__(self, **kwargs):
        """Init Base.

        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        :type platform: str

        :param version: Software version, default is "0".
        :type version: str

        Helpers
        :param uuid: Unique identifier.
        :type uuid: str

        :param note: Object description.
        :type note: Any
        """
        self._platform: str = h.init_platform(**kwargs)
        self._uuid: str = self._init_uuid(**kwargs)
        self.version: SwVersion = h.init_version(**kwargs)
        self.note: Any = self._init_note(**kwargs)

    def __repr__(self):
        """__repr__."""
        params = self._repr__params()
        kwargs = ", ".join(params)
        name = self.__class__.__name__
        return f"{name}({kwargs})"

    def __str__(self):
        """__str__."""
        return self.line

    @staticmethod
    def _init_uuid(**kwargs) -> str:
        """Init uuid."""
        if uuid := str(kwargs.get("uuid") or ""):
            return uuid
        return str(uuid1())

    @staticmethod
    def _init_note(**kwargs) -> Any:
        """Init note."""
        note = kwargs.get("note")
        if note is None:
            return ""
        return note

    # =========================== property ===========================

    @property
    def line(self) -> str:
        """Stub."""
        return ""

    @line.setter
    def line(self, line: str) -> None:
        """Stub."""
        _ = line  # noqa

    @property
    def platform(self) -> str:
        """Platform: Platform: "asa", "ios", "nxos"."""
        return self._platform

    @platform.setter
    def platform(self, platform: str) -> None:
        self._platform = h.init_platform(platform=platform)
        uuid = self.uuid
        self.line = self.line
        self.uuid = uuid

    @property
    def uuid(self) -> str:
        """Universally Unique Identifier."""
        return self._uuid

    @uuid.setter
    def uuid(self, uuid_: str) -> None:
        if not isinstance(uuid_, str):
            raise TypeError(f"{uuid_=} {str} expected")
        self._uuid = uuid_

    # =========================== method =============================

    def copy(self):
        """Copy the self object."""
        kwargs = self.data()
        return self.__class__(**kwargs)

    @abstractmethod
    def data(self, uuid: bool = False) -> DAny:
        """Convert self object to the dictionary.

        :param uuid: Return self.uuid in data.
        :type uuid: bool

        :return: The dictionary.
        """

    # =========================== helper =============================

    def _repr__add_param(self, param: str, params: LStr) -> LStr:
        """Add a param to the list of params."""
        if value := getattr(self, param):
            params.append(f"{param}={value!r}")
        return params

    def _repr__params(self) -> LStr:
        """Return parameters for __repr__."""
        params: LStr = [f"{self.line!r}"]
        if self._platform != IOS:
            params.append(f"platform={self._platform!r}")
        version = str(self.version)
        if version != "0":
            params.append(f"{version=!r}")
        params = self._repr__add_param("note", params)
        return params



"""ACE helper functions."""

import logging
import re
from functools import wraps
from ipaddress import IPv4Network
from string import ascii_letters, digits, punctuation
from time import time
from typing import Any, List, NamedTuple

from netports import SwVersion

from cisco_acl.types_ import LStr, StrInt, LInt, OInt, SInt, T2Str, T3Str, DInt, SStr, LIpNet

IOS = "ios"
MAX_LINE_LENGTH = 100
SEQUENCE_MAX = 4294967295
PLATFORMS = ("asa", "ios", "nxos")
ACTIONS = ("remark", "permit", "deny")
OPERATORS = ("eq", "gt", "lt", "neq", "range")

DEF_INDENT = "  "
OCTETS = r"\d+\.\d+\.\d+\.\d+"


# =============================== str ================================


def acl_help_to_name_port(output: str) -> DInt:
    """Transform `output` of ACL help to the dictionary where key is name, value is port.

    :param output: ACL help.
    :return: names and ports.
    :example:
        output: "bgp          Border Gateway Protocol (179)"
        return: {"bgp": 179}
    """
    data: DInt = {}
    lines = [s.strip().replace("(", " ") for s in output.split("\n")]
    for line in lines:
        name, port = findall2(r"(\S+).+\s(\d+)\)$", line)
        if name and port:
            data[name] = int(port)
    return data


def findall1(pattern: str, string: str, flags=0) -> str:
    """Parse 1st item of re.findall(). If nothing is found, returns an empty string.

    Group with parentheses in pattern is required.
    :return: Interested substring.
    :example:
        pattern: "a(b)cde"
        string: "abcde"
        return: "b"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [""])[0]
    if isinstance(result, str):
        return result
    if isinstance(result, tuple):
        return result[0]
    return ""


def findall2(pattern: str, string: str, flags=0) -> T2Str:
    """Parse 2 items of re.findall(). If nothing is found, returns 2 empty strings.

    Group with parentheses in pattern is required.
    :return: Two interested substrings.
    :example:
        pattern: "a(b)(c)de"
        string: "abcde"
        return: "b", "c"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [("", "")])[0]
    if isinstance(result, tuple) and len(result) >= 2:
        return result[0], result[1]
    return "", ""


def findall3(pattern: str, string: str, flags=0) -> T3Str:
    """Parse 3 items of re.findall(). If nothing is found, returns 3 empty strings.

    Group with parentheses in pattern is required.
    :return: Three interested substrings.
    :example:
        pattern: "a(b)(c)(d)e"
        string: "abcde"
        return: "b", "c", "d"
    """
    result = (re.findall(pattern=pattern, string=string, flags=flags) or [("", "", "")])[0]
    if isinstance(result, tuple) and len(result) >= 3:
        return result[0], result[1], result[2]
    return "", "", ""


def check_line_length(line: str) -> bool:
    """Return True if line length <= 100 chars, else raise ERROR."""
    length = len(line)
    expected = MAX_LINE_LENGTH
    if length > expected:
        raise ValueError(f"invalid {line=} {length=}, {expected=}")
    return True


def check_name(name: str) -> bool:
    """Check name.

    :return: True  - if first char is ascii_letters,
             False - chars can be punctuation, else raise ERROR.
    """
    if not name:
        raise ValueError(f"absent {name=}")
    skip_chas = {"?"}
    valid_chars: SStr = set(ascii_letters + digits + punctuation).difference(skip_chas)
    if invalid_chars := list(set(name).difference(valid_chars)):
        raise ValueError(f"{invalid_chars=} in {name=}")
    return True


def init_ace_action(action: str = "permit") -> str:
    """Init ACE action: "permit", "deny"."""
    action = str(action)
    expected = ["permit", "deny"]
    if action not in expected:
        raise ValueError(f"invalid {action=}, {expected=}")
    return action


def init_indent(**kwargs) -> str:
    """Init ACE lines indentation (default "  ")."""
    indent = kwargs.get("indent")
    if indent is None:
        indent = DEF_INDENT
    if not isinstance(indent, str):
        raise TypeError(f"{indent=} {str} expected")
    return indent


def init_line(line: str) -> str:
    """Init line, replace spaces to one space, checks length <= 100 chars."""
    if not isinstance(line, str):
        raise TypeError(f"{line=} {str} expected")
    line = replace_spaces(line)
    return line


def init_name(name: str) -> str:
    """ACL or address group name, without "ip access-list ".

    Requirements:
    - length <= 100 chars.
    - all chars are digits.
    - first char is ascii_letters, other chars are ascii_letters and punctuation.
    """
    if not name:
        name = ""
    if not isinstance(name, str):
        raise TypeError(f"{name=} {str} expected")
    name = name.strip()
    check_line_length(name)
    return name


def init_number(number: StrInt) -> str:
    """Init number, convert int to string."""
    return str(init_number_int(number))


def init_number_int(number: StrInt) -> int:
    """Init positive integer."""
    if isinstance(number, int):
        if number >= 0:
            return number
        raise ValueError(f"{number=} positive expected")

    if isinstance(number, str):
        if number.isdigit():
            return int(number)
        raise ValueError(f"{number=} digit expected")
    raise TypeError(f"{number=} {str} {int} expected")


# noinspection PyIncorrectDocstring
def init_platform(**kwargs) -> str:
    """Init device platform.

    :param platform: Not checked platform: "cisco_ios", "cisco_nxos", "cnx", "ios", "nxos".
    :return: Valid platform: "ios", "nxos".
    """
    platform = kwargs.get("platform") or ""
    if not platform:
        platform = IOS
    if not isinstance(platform, str):
        raise TypeError(f"{platform=} {str} expected")
    expected = sorted(["cisco_asa", "cisco_ios", "cisco_nxos", "cnx", *PLATFORMS])
    if platform not in expected:
        raise ValueError(f"Invalid {platform=}, {expected=}.")
    if platform in ["cisco_asa", "asa"]:
        return "asa"
    if platform in ["cisco_nxos", "cnx", "nxos"]:
        return "nxos"
    return IOS


# noinspection PyIncorrectDocstring
def init_protocol(line: str, **kwargs) -> str:
    """Init protocol, converts tcp, udp numbers to name.

    :param line: Protocol line. Example "eq www".
    :param protocol: Protocol name or number. Example: "tcp".
    """
    if line:
        protocol = str(kwargs.get("protocol") or "")
        if protocol == "6":
            protocol = "tcp"
        if protocol == "17":
            protocol = "udp"
        expected = ["", "tcp", "udp"]
        if protocol in expected:
            return protocol
    return ""


def init_remark_text(text: str) -> str:
    """Init Remark.text.

    :param text: text for remark.
    """
    if not isinstance(text, str):
        raise TypeError(f"{text=} {str} expected")
    text = text.strip()
    if not text:
        raise ValueError(f"{text=} value required")
    return text


# noinspection PyIncorrectDocstring
def init_type(**kwargs) -> str:
    """Init ACL type: "extended", "standard".

    :param platform: Platform: "asa", "ios", "nxos". Default "ios".
    :param type: Not checked ACL type: "extended", "standard", "ip access-list extended", etc.
    """
    platform = init_platform(**kwargs)
    _type = str(kwargs.get("type") or "").strip()
    if _type.startswith("ip access-list extended "):
        _type = "extended"
    if _type.startswith("ip access-list standard "):
        _type = "standard"
    if _type not in ["extended", "standard"]:
        _type = "standard"
        if platform == "nxos":
            _type = "extended"
    if platform == "nxos" and _type == "standard":
        expected = "extended"
        raise ValueError(f"invalid type={_type!r}, {expected=}")
    return _type


# noinspection PyIncorrectDocstring
def init_version(**kwargs) -> SwVersion:
    """Init version.

    Convert software version string to SwVersion object.
    :param version: software version string.
    :return: SwVersion object.
    """
    version = str(kwargs.get("version") or "")
    if not version:
        version = "0"
    return SwVersion(version)


def int_to_str(line: StrInt) -> str:
    """Init line, int or string convert to string, replace spaces."""
    if isinstance(line, int):
        if line < 0:
            raise ValueError(f"{line=} positive expected")
        line = str(line)
    if not isinstance(line, str):
        raise TypeError(f"{line=} {str} expected")
    line = replace_spaces(line)
    return line


def is_line_for_acl(line: str) -> bool:
    """Return True if line ready for ACL.

    If Starts with "allow", "deny", "remark" startswith` + `split` faster than `re`.
    """
    if line.startswith("permit "):
        return True
    if line.startswith("remark "):
        return True
    if line.startswith("deny "):
        return True

    digit, *items = line.split(" ", 1)
    if digit.isdigit() and items:
        return is_line_for_acl(items[0])
    return False


def lines_wo_spaces(line: str) -> LStr:
    r"""Split line by newline, replaces multiple white spaces with single space.

    :line: Text joined by "\n".
    :return: Lines with single spaces and split by "\n".
    :example:
        line: "  10 deny icmp any any  \n  20 permit ip any any  "
        return: ["10 deny icmp any any", "20 permit ip any any"]
    """
    if not isinstance(line, str):
        raise TypeError(f"{line=} {str} expected")
    lines = [replace_spaces(s) for s in line.split("\n")]
    return [s for s in lines if s]


def parse_remark_name(text: str, group_by: str) -> str:
    """Parse Rule name from Remark.text.

    :param text: Remark.text.
    :param group_by: Acl.group_by.
    :return: Rule name.
    :example:
        self.group_by = "= "
        parse_aceg_name("= C-1, text", "= ") -> "C-1"
    """
    name = text.split(",")[0]
    name = name.strip().replace(group_by, "")
    return name


def re_find_t(regex: str, line: str) -> tuple:
    """Find interested tuple.

    :param regex: Regex pattern, should contain groups.
    :param line: String body.
    :return: Tuple with interested values or empty tuple.
    """
    return (re.findall(regex, line) or [tuple()])[0]


def replace_spaces(line: str) -> str:
    """Replace multiple white spaces with single space."""
    return " ".join(line.split())


# =============================== int ================================


def init_int(line: StrInt) -> int:
    """Convert string to positive int."""
    if not isinstance(line, (int, str)):
        raise TypeError(f"{line=} {int} or {str} expected")
    if isinstance(line, str):
        line = line.strip()
        line = line.split()[0] if line else "0"
        if not line.isdigit():
            raise ValueError(f"{line=}, digit expected")
    line_: int = int(line)
    if line_ < 0:
        raise ValueError(f"{line_=}, positive expected")
    return line_


# =============================== list ===============================


def convert_to_lstr(items: Any) -> LStr:
    """Convert items to List[str].

    If items has other type, raise ERROR.
    """
    if not items:
        items = []
    if isinstance(items, str):
        items = [items]
    items = list(items)
    _items: LStr = []
    for item in items:
        if not isinstance(item, str):
            raise TypeError(f"{item=} {str} expected")
        _items.append(item)
    return _items


# ============================== ipnet ===============================


def prefix_to_ipnet(prefix: str) -> IPv4Network:
    """Convert prefix to ipnet, logging WARNING if invalid prefixlen.

    :param prefix: Prefix "A.B.C.D/LEN".
    :return: ipnet IPv4Network.
    """
    try:
        ipnet = IPv4Network(address=prefix)
    except ValueError as ex:
        if "has host bits set" not in str(ex):
            raise type(ex)(*ex.args)
        ipnet = IPv4Network(address=prefix, strict=False)
        msg = f"ValueError: {ex}, fixed to prefix {ipnet}"
        logging.warning(msg)
    return ipnet


def subnet_of(tops: LIpNet, bottoms: LIpNet) -> bool:
    """Check all IPv4Network  in`bottoms` ara subnets of any IPv4Network in `tops`.

    :param tops: List of IPv4Network in the top.
    :type bottoms: List of IPv4Network in the bottom.
    :return: True - if all `bottoms` atr subnets of `tops`.
    """
    if not (tops and bottoms):
        return False
    for bottom in bottoms:
        for top in tops:
            if bottom.subnet_of(top):
                break
        else:
            return False
    return True


# ============================== ports ===============================


class PortRange(NamedTuple):
    """TCP/UDP ports range."""

    string: str
    range: range
    min: int
    max: int


def ports_to_string(items: LInt) -> str:
    """Convert list of ports to string.

    :example:
        items: [1,3,4,5]
        return: "1,3-5"
    """
    if not items:
        return ""
    # if self.operator in ["eq", "neq"]:
    #     return ",".join([str(i) for i in items])

    items = sorted(items)
    ranges: LStr = []  # result
    item_1st: OInt = None
    for idx, item in enumerate(items, start=1):
        # not last iteration
        if idx < len(items):
            item_next = items[idx]
            if item_next - item <= 1:  # range
                if item_1st is None:  # start new range
                    item_1st = item
            else:  # int or end of range
                ranges.append(str(item) if item_1st is None else f"{item_1st}-{item}")
                item_1st = None
        # last iteration
        else:
            item_ = str(item) if item_1st is None else f"{item_1st}-{item}"
            ranges.append(item_)
    return ",".join(ranges)


def string_to_ports(ports: str) -> LInt:
    """Convert string to list of ports.

    :example:
        ports: "1,3-5"
        return: [1, 3, 4, 5]
    """
    values = [s for s in ports.split(",") if s]
    ints: SInt = {int(s) for s in values if s.isdigit()}

    ranges = set()
    for value in values:
        items = value.split("-")
        if len(items) == 2 and items[0].isdigit() and items[1].isdigit():
            ranges.add(value)

    ranges_t = _port_range_min_max(ranges)
    ports_calc = {i for o in ranges_t for i in o.range}  # ports in all ranges
    ports_calc.update(ints)
    ports_ = [i for i in ports_calc if 1 <= i <= 65535]
    return ports_


def _port_range_min_max(ranges) -> List[PortRange]:
    """Return named tuple of range_sting, range_int, port_min, port_max."""
    ranges_tup = []
    for range_string in ranges:
        port_min, port_max = map(int, range_string.split("-"))
        range_int = range(port_min, port_max + 1)
        ranges_tup.append(PortRange(range_string, range_int, port_min, port_max))
    ranges_tup = sorted(ranges_tup, key=lambda o: o.min)
    return ranges_tup


# ============================= wrapper ==============================


def check_start_step_sequence(method):
    """Wrap and check sequence numbers."""

    @wraps(method)
    def _wrapper(ace_o, start: int = 10, step: int = 10, **kwargs) -> int:
        """Check sequence numbers, max/min value for `start` and `step`.

        :param start: Starting sequence number. start=0 - delete all sequence numbers.
        :param step: Step to increment the sequence number.
        :param items: List of Ace objects (default self.items).
        :return: Last sequence number.
        :raises: ValueError if `start` or `step` does not match conditions.
        """
        if not 0 <= start <= SEQUENCE_MAX:
            raise ValueError(f"{start=} expected=0..{SEQUENCE_MAX}")
        if start and step < 1:
            raise ValueError(f"{step=} expected >= 1")
        if not start:
            step = 0

        sequence = method(ace_o, start, step, **kwargs)

        if sequence > SEQUENCE_MAX:
            raise ValueError(f"last {sequence=} expected=1..{SEQUENCE_MAX}")

        return sequence

    return _wrapper


def time_spent(func):
    """Wrap measure function execution time."""

    def wrap(*args, **kwargs):
        """Wrap."""
        started = time()
        _return = func(*args, **kwargs)
        elapsed = time() - started
        print(f"====== {func.__name__}, spent {elapsed:.3f}s ======")
        return _return

    return wrap



"""Parsing helpers."""

from cisco_acl import helpers as h
from cisco_acl.port_name import all_known_names
from cisco_acl.types_ import DStr, LStr


def parse_ace_extended(line: str) -> DStr:  # pylint: disable=too-many-locals
    """Parse extended ACE line to the dictionary.

    :param line: ACE string.
    :return: ACE dict.

    :example:
    parse_ace_extended("10 permit tcp host 1.1.1.1 eq 1025 10.0.0.0 255.0.0.0 eq web log") -> {
        "sequence": "10",
        "action": "permit",
        "protocol": "tcp",
        "srcaddr": "host 1.1.1.1",
        "srcport": "eq 1025",
        "dstaddr": "10.0.0.0 255.0.0.0",
        "dstport": "eq web",
        "option": "log",
     }
    """
    space = r"(?: )"
    text = r"\S+"
    addr = "|".join(
        [
            "any",  # "any"
            f"host {h.OCTETS}",  # "host A.B.C.D"
            f"(?:object-group|addrgroup) {text}",  # ios: "object-group", nxos: "addrgroup"
            h.OCTETS + r"/\d+",  # "A.B.C.D/LEN"
            f"{h.OCTETS} {h.OCTETS}",  # "A.B.C.D A.B.C.D"
        ]
    )

    re_sequence = r"(\d+)?"
    re_action = f"{space}?(permit|deny)"
    re_proto = f"({space}{text})?"
    re_srcaddr = f"{space}({addr})"
    re_srcport = "( .+)?"
    re_dstaddr = f"{space}({addr})"
    re_dstport = "( .+)?"

    regex = f"^{re_sequence}{re_action}{re_proto}{re_srcaddr}{re_srcport}{re_dstaddr}{re_dstport}"
    _items = h.re_find_t(regex, line)
    if not _items:
        return {}

    items = [s.strip() for s in _items]
    dstport_option = items[-1]
    result: DStr = _parse_dstport_option(dstport_option)
    data = dict(
        sequence=items[0],
        action=items[1],
        protocol=items[2],
        srcaddr=items[3],
        srcport=items[4],
        dstaddr=items[5],
        dstport=result["dstport"],
        option=result["option"],
    )
    return data


def parse_ace_standard(line: str) -> DStr:
    """Parse standard ACE line to the dictionary.

    :param line: ACE string.
    :return: ACE dict.

    :example:
        line: "10 permit host 1.1.1.1 log"
        return: {"sequence": "10",
                 "action": "permit",
                 "protocol": "ip",
                 "srcaddr": "host 1.1.1.1",
                 "srcport": "",
                 "dstaddr": "any",
                 "dstport": "",
                 "option": "log"}
    """
    space = r"(?: )"
    text = r"\S+"
    addr = "|".join(
        [
            "any",  # "any"
            f"host {h.OCTETS}",  # "host A.B.C.D"
            f"(?:object-group|addrgroup) {text}",  # ios: "object-group", nxos: "addrgroup"
            h.OCTETS + r"/\d+",  # "A.B.C.D/LEN"
            f"{h.OCTETS} {h.OCTETS}",  # "A.B.C.D A.B.C.D"
            h.OCTETS,  # host
        ]
    )

    re_sequence = r"(\d+)?"
    re_action = f"{space}?(permit|deny)"
    re_srcaddr = f"{space}({addr})"
    re_log = "( .+)?"

    regex = f"^{re_sequence}{re_action}{re_srcaddr}{re_log}"
    _items = h.re_find_t(regex, line)
    if not _items:
        return {}

    items = [s.strip() for s in _items]
    data = dict(
        sequence=items[0],
        action=items[1],
        protocol="ip",
        srcaddr=items[2],
        srcport="",
        dstaddr="any",
        dstport="",
        option=items[3],
    )
    return data


def _parse_dstport_option(line: str) -> DStr:
    """Split destination-ports and options based on end of ACE line.

    :param line: ACE string.
    :return: ACE dict.
    :example:
        line: "eq bgp www ack log"
        return: {"dstport": "eq bgp www", "option": "ack log"}
    """
    dstports: LStr = []  # result
    options: LStr = []  # result
    known_names = all_known_names()

    if items := line.split():
        operator = items[0]
        if operator in h.OPERATORS:
            dstports.append(operator)
            items = items[1:]
            for id_, item in enumerate(items):
                if item.isdigit() or item in known_names:
                    dstports.append(item)
                    continue
                options = items[id_:]
                break
        else:
            options.extend(items)
    return dict(dstport=" ".join(dstports), option=" ".join(options))


def parse_action(line: str) -> DStr:
    """Parse action from ACE line.

    :example:
        line: "10 remark TEXT"
        return: {"sequence": "10", "action": "remark", "text": "TEXT"}

    :example:
        line: "permit ip any any"
        return: "permit"
    """
    space = r"(?: )"
    actions = "|".join(h.ACTIONS)

    re_sequence = r"(\d+)?"
    re_action = f"{space}?({actions})"
    re_text = r"( .+)"
    regex = f"^{re_sequence}{re_action}{re_text}"
    _items = h.re_find_t(regex, line)
    if not _items:
        raise ValueError(f"invalid {line=}")
    items = [s.strip() for s in _items]
    data = dict(
        sequence=items[0],
        action=items[1],
        text=items[2],
    )
    return data


def parse_address(line: str) -> DStr:
    """Parse address item from address group.

    :example:
        line: "10 host 10.0.0.1"
        return: {"sequence": 10, "address": "host 10.0.0.1"}
    """
    re_sequence = r"(\d+\s+)?"
    re_address = "(.+)"
    regex = f"^{re_sequence}{re_address}"
    _items = h.re_find_t(regex, line)
    if not _items:
        raise ValueError(f"invalid {line=}")
    items = [s.strip() for s in _items]
    data = dict(
        sequence=items[0],
        address=items[1],
    )
    return data



"""Port - ACE TCP/UDP source or destination port object."""

from __future__ import annotations

from functools import total_ordering
from typing import List

from cisco_acl import helpers as h
from cisco_acl.base import Base
from cisco_acl.helpers import OPERATORS
from cisco_acl.port_name import PortName
from cisco_acl.types_ import LInt, LStr, IInt, DAny, StrInt


@total_ordering
class Port(Base):
    """Port - ACE TCP/UDP source or destination port object."""

    def __init__(self, line: str = "", **kwargs):
        """Init Port.

        :param line: TCP/UDP ports line.

        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        :type platform: str

        :param version: Software version, default is "0".
        :type version: str

        :param protocol: ACL protocol: "tcp", "udp", "".
        :type protocol: str

        Helpers
        :param note: Object description.
        :type note: Any

        :param port_nr: Well-known TCP/UDP ports as numbers.
            True  - all tcp/udp ports as numbers,
            False - well-known tcp/udp ports as names (default).
        :type port_nr: bool

        :example: ios, "eq" (can match multiple ports in single line).
            port = Port("eq www 443", platform="ios", protocol="tcp")
            port.line -> "eq www 443"
            port.operator -> "eq"
            port.items -> [80, 443]
            port.ports -> [80, 443]
            port.sport -> "80,443"

        :example: nxos, "neq" (can match only one port in single line).
            port = Port("neq www", platform="nxos", protocol="tcp")
            port.line -> "neq www"
            port.operator -> "neq"
            port.items -> [80]
            port.ports -> [1, 2, ..., 79, 81, ..., 65534, 65535]
            port.sport -> "1-79,81-65535"

        :example: Range.
            port = Port("range 1 3", platform="ios", protocol="tcp")
            port.line -> "range 1 3"
            port.operator -> "range"
            port.ports -> [1, 2, 3]
            port.sport == "1-3"
        """
        super().__init__(**kwargs)  # platform, note
        self._protocol = h.init_protocol(line=line, **kwargs)
        self._port_nr = bool(kwargs.get("port_nr"))
        self.line = line

    def __repr__(self):
        """__repr__."""
        params = self._repr__params()
        params = self._repr__add_param("protocol", params)
        kwargs = ", ".join(params)
        name = self.__class__.__name__
        return f"{name}({kwargs})"

    # ========================== redefined ===========================

    def __hash__(self) -> int:
        """__hash__."""
        return self.line.__hash__()

    def __eq__(self, other) -> bool:
        """== equality."""
        return self.__hash__() == other.__hash__()

    def __lt__(self, other) -> bool:
        """< less than."""
        if self.__class__ == other.__class__:
            if self._operator and other.operator:
                if self._operator != other.operator:
                    return self._operator < other.operator
                if self._items[0] != other.items[0]:
                    return self._items[0] < other.items[0]
                if self._items[-1] != other.items[-1]:
                    return self._items[-1] < other.items[-1]
        return False

    # =========================== property ===========================

    @property
    def items(self) -> LInt:
        """ACE TCP/UDP list of int protocol numbers.

        :return: List of ports.
        :example:
            Port("eq www 443") -> [80, 443]
            Port("neq www") -> [80]
        """
        return self._items

    @items.setter
    def items(self, items: IInt) -> None:
        if not isinstance(items, (set, list, tuple)):
            raise TypeError(f"{items=} {list} expected")
        items_ = [str(i) for i in list(items)]
        items_.insert(0, self._operator)
        self.line = " ".join(items_)

    @property
    def line(self) -> str:
        """ACE source or destination TCP/UDP ports."""
        if not (self._operator and self._items and self._protocol in ["tcp", "udp"]):
            return ""
        if self._port_nr:
            items_s = " ".join([str(i) for i in self._items])
            return f"{self._operator} {items_s}"
        port_name = PortName(protocol=self._protocol, platform=self._platform, version=self.version)
        data = port_name.ports()
        items_s = " ".join([str(data.get(i) or i) for i in self._items])
        return f"{self._operator} {items_s}"

    @line.setter
    def line(self, line: str) -> None:
        line = h.init_line(line)
        items = line.split()
        if not items:
            self._operator = ""
            self._items = []
            self._ports = []
            self._sport = ""
            return

        self._operator = self._line__operator(items)
        items = items[1:]
        _items: LInt = self._line__items_to_ints(items)
        ports: LInt = self._items_to_ports(_items)
        self._items = _items
        self._ports = ports
        self._sport = h.ports_to_string(ports)

    @property
    def operator(self) -> str:
        """ACE TCP/UDP port operator: "eq", "gt", "lt", "neq", "range".

        :return: Operator.
        :example:
            Port("eq www 443") -> "eq"
        """
        return self._operator

    @property
    def port_nr(self) -> bool:
        """Return well-known TCP/UDP ports as numbers.

        True  - all tcp/udp ports as numbers,
        False - well-known tcp/udp ports as names (default).
        """
        return self._port_nr

    @port_nr.setter
    def port_nr(self, port_nr: bool) -> None:
        self._port_nr = bool(port_nr)

    @property
    def ports(self) -> LInt:
        """ACE list of int TCP/UDP port numbers.

        :example:
            Port("eq www 443") -> [80, 443]
            Port("neq www") -> [1, 2, ..., 79, 81, ..., 65534, 65535]
        """
        return self._ports

    @ports.setter
    def ports(self, ports: IInt) -> None:
        if not isinstance(ports, (set, list, tuple)):
            raise TypeError(f"{ports=} {list} expected")
        ports = list(ports)
        items_ = self._ports_to_items(ports)
        items = [str(i) for i in items_]
        items.insert(0, self.operator)
        self.line = " ".join(items)

    @property
    def protocol(self) -> str:
        """Protocol name: "tcp", "udp", ""."""
        return self._protocol

    @protocol.setter
    def protocol(self, protocol: StrInt) -> None:
        if not isinstance(protocol, (int, str)):
            raise TypeError(f"{protocol=} {int} {str} expected")
        self._protocol = h.init_protocol(line=self.line, protocol=protocol)
        uuid = self.uuid
        self.line = self.line
        self.uuid = uuid

    @property
    def sport(self) -> str:
        """ACE string of TCP/UDP ports range.

        :example:
            Port("eq 1 3 4 5") -> "1,3-5"
        """
        return self._sport

    @sport.setter
    def sport(self, sport: str) -> None:
        if not isinstance(sport, str):
            raise TypeError(f"{sport=} {str} expected")
        self.ports = h.string_to_ports(sport)

    # =========================== method =============================

    def data(self, uuid: bool = False) -> DAny:
        """Convert Port object to the dictionary.

        :param uuid: Return self.uuid in data.
        :return: Port data.
        :example:
            address = Address("10.0.0.0/24", platform="nxos")
            address.data() -> {
                "line": "10.0.0.0/24",
                "platform": "nxos",
                "version": "0",
                "items": [],
                "note": "",
                "addrgroup": "",
                "ipnet": IPv4Network("10.0.0.0/24"),
                "prefix": "10.0.0.0/24",
                "subnet": "10.0.0.0 255.255.255.0",
                "wildcard": "10.0.0.0 0.0.0.255",
            }
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            version=str(self.version),
            note=self.note,
            protocol=self._protocol,
            port_nr=self._port_nr,
            # property
            items=self._items,
            operator=self._operator,
            ports=self._ports,
            sport=self._sport,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data

    # =========================== helper =============================

    @staticmethod
    def _line__operator(items: LStr) -> str:
        """Get operator from items.

        :example:
            self._line__operator(["eq", "www"]) -> ["eq"]
        """
        operator, *items = items
        expected: LStr = list(OPERATORS)
        if operator not in expected:
            raise ValueError(f"invalid port {operator=}, {expected=}")
        return operator

    def _line__items_to_ints(self, items: LStr) -> LInt:
        """Convert named and digit items to int items.

        :example:
            self._line__items_to_ints(["www", "443"]) -> [80, 443]
        """
        if not items:
            raise ValueError(f"absent ports={items}")

        # convert to int
        ports: LInt = []  # result
        for item in items:
            if item.isdigit():
                ports.append(int(item))
                continue
            port_name = PortName(
                protocol=self._protocol, platform=self._platform, version=self.version
            )
            data = port_name.names()
            if port_nr := data.get(item):
                ports.append(port_nr)
                continue
            msg = f"invalid {item=}"
            if expected := sorted(data):
                msg = f"{msg}, {expected=}"
            raise ValueError(msg)

        # validation
        operator = self._operator
        if operator in ["lt", "gt"] and len(ports) != 1:
            raise ValueError(f"invalid {operator=} with {ports=}")
        if operator == "range" and len(ports) != 2:
            raise ValueError(f"invalid {operator=} with {ports=} expected 2 ports")
        if self._operator in ["eq", "neq"]:
            platform = self._platform
            if platform in ["asa", "nxos"] and len(ports) != 1:
                raise ValueError(f"invalid count of {ports=}, for {platform=} expected 1 port")

        return sorted(ports)

    def _items_to_ports(self, items: LInt) -> LInt:
        """Transform line items to ports.

        :example:
            self.operator == "range"
            self._items_to_ports([1, 4]) ->[1, 2, 3, 4]

            self.operator == "ge"
            self._items_to_ports([4, 5, 6, ..., 65535]) -> [4]
        """
        operator = self._operator
        if operator == "eq":
            return items
        if operator == "range":
            items = list(range(items[0], items[-1] + 1))
            return items

        all_ports = list(range(1, 65535 + 1))
        if operator == "neq":
            items = [i for i in all_ports if i not in items]
            return items
        if operator == "gt":
            items = [i for i in all_ports if i > items[0]]
            return items
        if operator == "lt":
            items = [i for i in all_ports if i < items[0]]
            return items
        raise ValueError(f"invalid port {operator=}")

    def _ports_to_items(self, ports: LInt) -> LInt:
        """Transform ports to line items.

        :example:
            self.operator == "range"
            self._ports_to_items([1, 2, 3, 4]) -> [1, 4]

            self.operator == "ge"
            self._ports_to_items([4, 5, 6, ..., 65535]) -> [4]
        """
        operator = self._operator
        if operator == "eq":
            return ports
        if operator == "range":
            return [ports[0], ports[-1]]
        if operator == "neq":
            items: LInt = list(range(1, 65535 + 1))
            for port in ports:
                items.remove(port)
            return items
        if operator == "gt":
            return [ports[0] - 1]
        if operator == "lt":
            return [ports[1] + 1]
        raise ValueError(f"invalid port {operator=}")


LPort = List[Port]



"""Wildcard network that of Cisco ACL."""

from __future__ import annotations

from functools import lru_cache
from ipaddress import NetmaskValueError, IPv4Address, IPv4Network
from itertools import product

from cisco_acl import helpers as h
from cisco_acl.base import Base
from cisco_acl.types_ import LIpNet, LInt, DAny, OIpNet, T2IpAddr, TLintInt

PREFIX_LEN = 32  # IPv4 prefix length
ALL_ONES = (2**PREFIX_LEN) - 1
DEF_NCWB = 16  # Default count of non-contiguous wildcard bits
MAX_NCWB = 30  # Maximum allowed count of non-contiguous wildcard bits


class Wildcard(Base):
    """Wildcard network that of Cisco ACL."""

    def __init__(self, line: str, **kwargs):
        """Init Wildcard.

        :param line: Network with wildcard mask.
        :param max_ncwb: Max count of non-contiguous wildcard bits. Allowed in range 0..30.
            0  - contiguous wildcard, 1 prefix
            30 - max allowed, 1073741824 prefixes
            16 - default, 65536 prefixes
        :type max_ncwb: int

        Helpers
        :param uuid: Unique identifier.
        :type uuid: str

        :param note: Object description.
        :type note: Any

        :raises NetmaskValueError: If non-contiguous wildcard increase max_ncwb.
        """
        self.ipnet: OIpNet = None  # IPv4Network of contiguous wildcard
        self._ncwb: LInt = []  # non-contiguous wildcard bits
        self._prefixlen: int = 0  # Prefix length of contiguous wildcard
        super().__init__(**kwargs)  # platform, note
        self.max_ncwb: int = init_max_ncwb(**kwargs)
        self.line = line

    def __repr__(self):
        """__repr__."""
        params = self._repr__params()

        max_ncwb = self.max_ncwb
        if max_ncwb != DEF_NCWB:
            params.append(f"{max_ncwb=!r}")

        name = self.__class__.__name__
        kwargs = ", ".join(params)
        return f"{name}({kwargs})"

    def __str__(self):
        """__str__."""
        return self.line

    # =========================== property ===========================

    @property
    def line(self) -> str:
        """Line of network with wildcard mask."""
        return f"{self._prefix} {self._wildmask}"

    @line.setter
    def line(self, line: str) -> None:
        line = h.init_line(line)
        prefix_o, wildmask_o = self._create_prefix(line)
        self._prefix = prefix_o
        self._wildmask = wildmask_o
        self.ipnet = self._create_ipnet()
        ncwb, prefixlen = self._create_ncwb()
        self._ncwb = ncwb
        self._prefixlen = prefixlen

    @property
    def max_ncwb(self) -> int:
        """Max count of non-contiguous wildcard bits."""
        return self._max_ncwb

    @max_ncwb.setter
    def max_ncwb(self, max_ncwb: int) -> None:
        """Max count of non-contiguous wildcard bits.

        :param max_ncwb: Max count of non-contiguous wildcard bits. Allowed in range 0..30.
            0  - contiguous wildcard, 1 prefix
            30 - max allowed, 1073741824 prefixes
            16 - default, 65536 prefixes
        """
        self._max_ncwb = init_max_ncwb(max_ncwb=max_ncwb)

    @property
    def prefix(self) -> str:
        """Wildcard network."""
        return str(self._prefix)

    @property
    def wildmask(self) -> str:
        """Wildcard mask."""
        return str(self._wildmask)

    # =========================== classmethod ============================

    # noinspection PyIncorrectDocstring
    @classmethod
    def fprefix(cls, prefix: str, **kwargs) -> Wildcard:
        """Convert prefix to Wildcard object.

        :param prefix: Prefix "A.B.C.D/LEN".
        :type prefix: str

        :param max_ncwb: Max count of non-contiguous wildcard bits.
        :type max_ncwb: int

        :return: Wildcard object
        """
        ipnet: IPv4Network = h.prefix_to_ipnet(prefix)
        line: str = ipnet.with_hostmask.replace("/", " ")
        return Wildcard(line, **kwargs)

    # noinspection PyIncorrectDocstring
    @classmethod
    def fsubnet(cls, subnet: str, **kwargs) -> Wildcard:
        """Convert subnet to Wildcard object.

        :param subnet: Subnet with mask "A.B.C.D A.B.C.D".
        :type subnet: str

        :param max_ncwb: Max count of non-contiguous wildcard bits.
        :type max_ncwb: int

        :return: Wildcard object.
        :rtype: Wildcard
        """
        items = subnet.split()
        if len(items) != 2:
            raise ValueError(f"invalid {subnet=}")
        subnet_, mask = items
        if mask == "0.0.0.0":
            wildmask = invert_mask(mask)
            wildcard = f"{subnet_} {wildmask}"
            return Wildcard(wildcard, **kwargs)

        subnet = subnet.replace(" ", "/")
        ipnet = IPv4Network(subnet)
        if subnet != ipnet.with_netmask:
            raise ValueError(f"{subnet=} not equal ipnet={str(ipnet)!r}")
        wildcard = ipnet.with_hostmask.replace("/", " ")
        return Wildcard(wildcard, **kwargs)

    # =========================== method =============================

    def data(self, uuid: bool = False) -> DAny:
        """Convert Wildcard object to the dictionary.

        :param uuid: Return self.uuid in data.
        :type uuid: bool

        :return: The dictionary.
        """
        data = dict(
            # init
            line=self.line,
            max_ncwb=self.max_ncwb,
            platform=self._platform,
            version=str(self.version),
            note=self.note,
            # property
            ipnet=self.ipnet,
            prefix=self.prefix,
            wildmask=self.wildmask,
        )
        if uuid:
            data["uuid"] = self.uuid
        return data

    @lru_cache
    def ipnets(self) -> LIpNet:
        """List of IPv4Network that match this wildcard.

        :return: List of IPv4Network.
        :example:
            wildcard = Wildcard("10.0.0.0 0.0.1.3")
            wildcard.ipnets() -> [IPv4Network("10.0.0.0/30"),
                                  IPv4Network("10.0.1.0/30")]
        """
        ipnets: LIpNet = []
        prefix_i = int(self._prefix)
        repeat = len(self._ncwb)
        for bits_values in product((0, 1), repeat=repeat):
            prefix_i_ = prefix_i
            for idx, value in zip(self._ncwb, bits_values):
                mask = 1 << idx
                if value:
                    prefix_i_ |= mask
                else:
                    prefix_i_ &= ~mask
            ipnet = IPv4Network((prefix_i_, self._prefixlen))
            ipnets.append(ipnet)
        return ipnets

    # =========================== helper =============================

    def _create_ipnet(self) -> OIpNet:
        """Init ipnet."""
        prefix = str(self._prefix)
        wildmask = str(self._wildmask)
        try:
            if str(wildmask) == "255.255.255.255":
                subnet = f"{prefix}/0"
            elif str(wildmask) == "0.0.0.0":
                subnet = f"{prefix}/{PREFIX_LEN}"
            else:
                mask = invert_mask(wildmask)
                if not is_mask(mask):
                    mask = ""
                subnet = f"{prefix}/{mask}"
            ipnet = IPv4Network(subnet)
        except NetmaskValueError:
            ipnet = None
        return ipnet

    def _create_ncwb(self) -> TLintInt:
        """Init non-contiguous wildcard bits and prefix length.

        :return: List of non-contiguous wildcard bits, prefixlen.
        :example:
            _init_prefix("10.0.0.0 0.0.0.3") -> IPv4Address("10.0.0.0"), IPv4Address("0.0.0.3")
        """
        wild_bits: LInt = [int(b) for b in format(int(self._wildmask), f"0{PREFIX_LEN}b")]
        wb_idxs: LInt = [i for i, e in enumerate(reversed(wild_bits)) if e == 1]
        prefixlen_idx: int = self._prefixlen_idx(wb_idxs)
        ncwb: LInt = self._ncw_bits(wb_idxs, prefixlen_idx)
        prefixlen: int = PREFIX_LEN - prefixlen_idx
        return ncwb, prefixlen

    @staticmethod
    def _create_prefix(line: str) -> T2IpAddr:
        """Convert line to prefix, wildmask as IPv4Address.

        :return: prefix, wildmask.
        :example:
            _init_prefix("10.0.0.0 0.0.0.3") -> IPv4Address("10.0.0.0"), IPv4Address("0.0.0.3")
        """
        items = line.split()
        if len(items) != 2:
            raise ValueError(f"invalid {line=}")
        prefix_i = int(IPv4Address(items[0]))
        wildmask_o = IPv4Address(items[1])
        wildmask_i = int(wildmask_o)
        inverted_i = ALL_ONES ^ wildmask_i
        prefix_i = prefix_i & inverted_i
        prefix_o = IPv4Address(prefix_i)
        return prefix_o, wildmask_o

    def _ncw_bits(self, wb_idxs: LInt, prefixlen_idx: int) -> LInt:
        """Return non-contiguous wildcard mask bits.

        :param wb_idxs: Wildcard mask bit indexes.
        :param prefixlen_idx: Index to get indexes of IPv4Network prefix length.
        :raise NetmaskValueError: If non-contiguous wildcard mask bits count increase allowed.
        """
        ncwb: LInt = wb_idxs[prefixlen_idx:]
        ncwb.reverse()

        count = len(ncwb)
        if count > self.max_ncwb:
            wild_mask = str(self._wildmask)
            max_ncwb = self.max_ncwb
            msg = f"non-contiguous wildcard bits {count=} increases {max_ncwb=} in {wild_mask=}"
            raise NetmaskValueError(msg)

        return ncwb

    @staticmethod
    def _prefixlen_idx(mask_bit_idxs: LInt) -> int:
        """Get last bits in wildcard mask amd returns index to get prefix length.

        :return: Index to get indexes of IPv4Network prefix length.
        """
        prefixlen_idx = 0
        for idx, mask_bit_idx in enumerate(mask_bit_idxs):
            if idx != mask_bit_idx:
                break
            prefixlen_idx += 1
        return prefixlen_idx


# ============================ functions =============================


# noinspection PyIncorrectDocstring
def init_max_ncwb(**kwargs) -> int:
    """Init max non-contiguous wildcard bits count.

    :param max_ncwb: Max count of non-contiguous wildcard bits. Allowed in range 0..30.
        0  - contiguous wildcard, 1 prefix
        30 - max allowed, 1073741824 prefixes
        16 - default, 65536 prefixes
    """
    max_ncwb = kwargs.get("max_ncwb")
    if max_ncwb is None:
        max_ncwb = DEF_NCWB
    if not isinstance(max_ncwb, int):
        raise TypeError(f"{max_ncwb=} {int} expected")
    if not 0 <= max_ncwb <= MAX_NCWB:
        raise ValueError(f"invalid {max_ncwb=}, allowed in range 0..{MAX_NCWB}")
    return max_ncwb


def invert_mask(mask: str) -> str:
    """Invert mask to wildcard and vice versa.

    :param mask: Mask or wildmask "0.0.0.3".
    :return: Inverted mask or wildmask.
    :example: wildmask to mask.
        invert_mask("0.0.0.3") -> "0.0.0.252"
    :example: mask to wildmask.
        invert_mask("0.0.0.252") -> "0.0.0.3"
    """
    return ".".join([str(255 - int(s)) for s in mask.split(".")])


def is_contiguous_wildmask(mask: str) -> bool:
    """Check mask.

    :return: True - if contiguous wildcard mask,
             False - if non-contiguous wildcard mask or generic mask.
    :example:
        is_contiguous_wildmask("0.0.0.3") -> True
        is_contiguous_wildmask("0.0.3.3") -> False
        is_contiguous_wildmask("255.255.255.252") -> False
    """
    if mask_i := sum_octets(mask):
        bits = f"{mask_i:b}"
        return "0" not in set(bits)
    return True


def is_mask(mask: str) -> bool:
    """Check mask.

    :return: True - if generic mask, False - if wildcard mask.
    :example:
    is_mask("255.255.255.252") -> True
        is_mask("0.0.0.3") -> False
        is_mask("0.0.3.3") -> False
    """
    if mask_i := sum_octets(mask):
        bits = "".join([str(b) for b in format(int(mask_i), f"0{PREFIX_LEN}b")])
        bits = bits.lstrip("1")
        return "1" not in set(bits)
    return True


def sum_octets(mask: str) -> int:
    """Return sum of "A.B.C.D" octets."""
    octets: LInt = [int(s) for s in mask.split(".")]
    if len(octets) != 4:
        raise ValueError(f"invalid {mask=}, expected 4 octets")
    return sum([octets[0] * 256**3, octets[1] * 256**2, octets[2] * 256, octets[3]])



"""Group of Items."""

from __future__ import annotations

from typing import Iterable


class Group:
    """Group of Items."""

    def __init__(self, items=None):
        """Init Group."""
        self.items = []
        if items:
            self.items = list(items)

    # ======================= special methods ========================

    def __add__(self, other) -> Group:
        """Return self.items + other.items."""
        group = Group(self.items.copy())
        group.update(other.items)
        return group

    def __contains__(self, item) -> bool:
        """Return key in self."""
        return item in self.items

    def __delitem__(self, idx: int) -> None:
        """Delete self.items[key]."""
        self.items.__delitem__(idx)

    def __getitem__(self, idx: int):
        """__getitem__."""
        return self.items[idx]

    def __iter__(self):
        """Iterate."""
        return iter(self.items)

    def __len__(self) -> int:
        """__len__."""
        return len(self.items)

    def __reversed__(self):
        """Return a reverse iterator over the list."""
        for item in self.items[::-1]:
            yield item

    # =========================== method =============================

    def add(self, item) -> None:
        """Add new item to self.items list, if it is not in self.items."""
        if item not in self.items:
            self.items.append(item)

    def append(self, item) -> None:
        """Append item to the end of the self.items list."""
        self.items.append(item)

    def clear(self) -> None:
        """Remove all items from the self.items list."""
        self.items = []

    def copy(self):
        """Return a shallow copy of the self.items list."""
        return self.items.copy()

    def count(self, item):
        """Return number of occurrences of the self.items."""
        return self.items.count(item)

    def delete(self, item) -> None:
        """Remove item from the self.items list."""
        if item in self.items:
            self.items.remove(item)

    def extend(self, items: Iterable) -> None:
        """Extend the self.items list by appending items."""
        if isinstance(items, (list, set, tuple)):
            self.items.extend(list(items))
            return
        raise TypeError(f"{items=} {list} expected")

    def index(self, *args) -> int:
        """Return first index of item.

        Raise ValueError if the value is not present.
        """
        return self.items.index(*args)

    def insert(self, *args) -> None:
        """Insert item before index."""
        return self.items.insert(*args)

    def pop(self, *args):
        """Remove and return item at index (default last).

        Raise IndexError if list is empty or index is out of range.
        """
        return self.items.pop(*args)

    def remove(self, *args) -> None:
        """Remove first occurrence of items in the self.items.

        Raise ValueError if the item is not present.
        """
        self.items.remove(*args)

    def reverse(self) -> None:
        """Reverse order of items in the self.items list."""
        self.items.reverse()

    def sort(self, *args, **kwargs) -> None:
        """Sort the self.items list in ascending order.

        :example:
            self.items.sort(reverse=True|False, key=myFunc)
        """
        self.items.sort(*args, **kwargs)

    def update(self, items: list) -> None:
        """Extend list by adding items to self.items list, if it is not in the self.items."""
        for item in items:
            self.add(item)



"""AddrGroup.

Group of AddressAg addresses configured in "object-group network" (ios) or
"object-group ip address" (nxos).
"""

from __future__ import annotations

import logging
from functools import total_ordering
from ipaddress import IPv4Network
from typing import Any, Dict, List, Union

from cisco_acl import parsers, helpers as h
from cisco_acl.address_ag import AddressAg, OAddressAg, LUSAddressAg
from cisco_acl.address_ag import LAddressAg
from cisco_acl.base import Base
from cisco_acl.group import Group
from cisco_acl.types_ import LStr, LIpNet, DAny
from cisco_acl.wildcard import init_max_ncwb


@total_ordering
class AddrGroup(Base, Group):
    """AddrGroup."""

    def __init__(self, line: str = "", **kwargs):
        r"""Init AddrGroup.

        :param line: Address group config line.
        :type line: str

        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        :type platform: str

        :param version: Software version, default is "0".
        :type version: str

        Helpers
        :param note: Object description.
        :type note: Any

        :param indent: Address lines indentation (default "  ").
        :type indent: str

        :param max_ncwb: Max count of non-contiguous wildcard bits.
        :type max_ncwb: int

        Alternate way to get `name` and `items`, if `line` absent.
        :param name: Address group name (default from `line`)
        :type name: str

        :param items: List of addresses in group.
        :type items: List[str], List[AddressAg]

        :example:
            address = AddrGroup("object-group ip address NAME\nhost 10.0.0.1")
            result:
                address.line == "object-group ip address NAME\n  host 10.0.0.1"
                address.platform == "ios"
                address.indent == "  "
                address.items == [AddressAg("host 10.0.0.1")]
        """
        self._line: str = ""
        self._name = ""
        self._items: LAddressAg = []
        Base.__init__(self, **kwargs)  # platform, note
        Group.__init__(self)
        self._indent = h.init_indent(**kwargs)
        self.max_ncwb: int = init_max_ncwb(**kwargs)
        if name := str(kwargs.get("name") or ""):
            self.name = name
        if items := kwargs.get("items") or []:
            self.items = items
            return
        self.line = line

    def __hash__(self) -> int:
        """__hash__."""
        return self._name.__hash__()

    def __eq__(self, other) -> bool:
        """== equality."""
        if self.__class__ == other.__class__:
            return self.__hash__() == other.__hash__()
        return False

    def __lt__(self, other) -> bool:
        """< less than."""
        if self.__class__ == other.__class__:
            return self._name < other.name
        return False

    def __contains__(self, other: UAddrGr) -> bool:
        """__contains__."""
        if isinstance(other, AddressAg):
            if other in self._items:
                return True
            for item in self._items:
                if not isinstance(item, AddressAg):
                    raise TypeError(f"{item=} {AddressAg} expected")
                if other in item:
                    return True
            return False

        if isinstance(other, AddrGroup):
            for other_item in other.items:
                if not isinstance(other_item, (AddressAg, AddrGroup)):
                    raise TypeError(f"{other_item=} {AddressAg} expected")
                if other_item in self._items:
                    return True
                for item in self._items:
                    if not isinstance(item, AddressAg):
                        raise TypeError(f"{item=} {AddressAg} expected")
                    if other in item:
                        return True
            return False
        raise TypeError(f"{other=} {UAddrGr} expected")

    # =========================== property ===========================

    @property
    def indent(self) -> str:
        """Address lines indentation (default  "  ")."""
        return self._indent

    @indent.setter
    def indent(self, indent: Any) -> None:
        self._indent = h.init_indent(indent=indent)

    @property
    def items(self) -> LAddressAg:
        """List of AddressAg objects."""
        return self._items

    @items.setter
    def items(self, items: LUSAddressAg) -> None:
        if isinstance(items, (str, AddressAg, AddrGroup)):
            items = [items]
        if not isinstance(items, (list, tuple)):
            raise TypeError(f"{items=} {list} expected")

        _items: LAddressAg = []
        for item in items:
            if isinstance(item, (AddressAg, AddrGroup)):
                item._platform = self._platform
                item.version = self.version
                _items.append(item)
            elif isinstance(item, dict):
                item["platform"] = self._platform
                item["version"] = str(self.version)
                addr_o = AddressAg(**item)
                _items.append(addr_o)
            elif isinstance(item, str):
                line = h.init_line(item)
                # description
                if item.startswith("description "):  # todo description
                    continue
                # AddressAg
                item_: OAddressAg = self._line_to_address(line)
                if not item_:
                    msg = f"invalid {line=}"
                    logging.warning(msg)
                    continue
                _items.append(item_)
            else:
                raise TypeError(f"{item=} {str} expected")
        self._items = _items

    @property
    def line(self) -> str:
        r"""Address group config line.

        :example:
            self: AddrGroup("object-group ip address NAME\nhost 10.0.0.1")
            return: "object-group ip address NAME\n  host 10.0.0.1"
        """
        items = [f"{self._indent}{o.line}" for o in self._items]
        line = "\n".join([self.cmd_addgr_name(), *items])
        return line

    @line.setter
    def line(self, line: str) -> None:
        items = h.lines_wo_spaces(line)
        if not items:
            if self.name and self.items:
                return
            raise ValueError(f"absent {line=}")

        name = ""
        item1, *items = items
        if self._platform == "nxos":
            name = h.findall1(r"^object-group ip address (.+)", item1)
        elif self._platform == "ios":
            name = h.findall1(r"^object-group network (.+)", item1)
        if not name:
            raise ValueError(f"absent {name=} in {line=}")
        if len(name.split()) > 1 or not name:
            raise ValueError(f"invalid {name=} in {line=}")
        self.name = name

        if not items and not self.items:
            raise ValueError(f"absent {items=} in {line=}")
        addresses: LAddressAg = []
        re_idx, re_address = r"(\d+(?:\s+))?", "(.+)"
        regex = f"{re_idx}{re_address}"
        for item in items:
            idx, item = h.findall2(regex, item)
            try:
                address = AddressAg(line=item, platform=self._platform)
            except ValueError:
                msg = f"invalid {item=}"
                logging.debug(msg)
                continue
            address.sequence = h.init_int(idx)
            addresses.append(address)
        if not addresses:
            raise ValueError(f"absent {addresses=} in {line=}")
        self.items = addresses

    @property
    def name(self) -> str:
        """Address group name."""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        if name := h.init_name(name):
            h.check_name(name)
        self._name = name

    @property
    def platform(self) -> str:
        """Platform: Platform: "asa", "ios", "nxos"."""
        return self._platform

    @platform.setter
    def platform(self, platform: str) -> None:
        """Change platform, normalizes self.items regarding the new platform.

        :param platform: Platform: "asa", "ios", "nxos". Default "ios".
        """
        self._platform = h.init_platform(platform=platform)

        for item in self._items:
            item.platform = self._platform

        data = self.data(uuid=True)
        self.__init__(**data)  # type: ignore

    # =========================== method =============================

    def data(self, uuid: bool = False) -> DAny:
        """Return AddrGroup data as dict.

        :param uuid: Return self.uuid in data.
        :type uuid: bool

        :return: Address group data.
        """
        data = dict(
            # init
            line=self.line,
            platform=self._platform,
            version=str(self.version),
            note=self.note,
            indent=self._indent,
            name=self._name,
            items=[o.data(uuid=uuid) for o in self._items],
        )
        if uuid:
            data["uuid"] = self.uuid
        return data

    def cmd_addgr_name(self) -> str:
        """Address group name line, with "object-group ip address" keyword in beginning.

        :return: Address group name line.

        :example:
            self.name: "NAME"
            self.platform: "ios"
            return: "object-group network NAME"

        :example:
            self.name: "NAME"
            self.platform: "nxos"
            return: "object-group ip address NAME"
        """
        if self._platform == "nxos":
            return f"object-group ip address {self._name}"
        # ios
        return f"object-group network {self._name}"

    def ipnets(self) -> LIpNet:
        """List of IPv4Network from all addresses in address group.

        return: List of IpNetwork.
        :raises ValueError: If one of the address is non-contiguous wildcard.

        :example: all items ara valid addresses
            self.items: [AddressAd("10.0.0.0/30"),
                         AddressAd("object-group ip address NAME")]  # 10.1.1.0/30
            return: [IpNetwork("10.0.0.0/30"), IpNetwork("10.1.1.0/30")]

        :example: non-contiguous wildcard
            self.items: [AddressAd("10.0.0.0 0.0.3.3")]
            raises: ValueError
        """
        ipnets: LIpNet = []
        for address in self._items:
            ipnet = address.ipnet
            if not isinstance(ipnet, IPv4Network):
                raise TypeError(f"{self.line} {ipnet=} {IPv4Network} expected")
            ipnets.append(ipnet)
        return ipnets

    def prefixes(self) -> LStr:
        """Prefixe from all addresses in address group.

        :return: Prefixes "A.B.C.D/LEN".

        :example:
            self.items: [AddressAd("10.0.0.0/30"),
                         AddressAd("object-group ip address NAME")]  # 10.1.1.0/30
            return: [IpNetwork("10.0.0.0/30"), IpNetwork("10.1.1.0/30")]
        """
        ipnets = self.ipnets()
        return [str(o) for o in ipnets]

    # noinspection PyIncorrectDocstring
    @h.check_start_step_sequence
    def resequence(self, start: int = 10, step: int = 10, **kwargs) -> int:
        """Change sequence numbers for all addresses in address group.

        :param start: Starting sequence number. start=0 - delete all sequence numbers.
        :param step: Step to increment the sequence number.
        :param items: List of AddressAg objects (default self.items).
        :return: Last sequence number.
        """
        items: LAddressAg = kwargs.get("items") or self._items
        sequence: int = int(start)
        count = len(items)

        for id_, item in enumerate(items, start=1):
            item.sequence = sequence
            if id_ < count:
                sequence += step
        return sequence

    def subnets(self) -> LStr:
        """Subnets from all addresses in address group.

        :return: Subnets with mask "A.B.C.D A.B.C.D".

        :example:
            self.items: [AddressAd("10.0.0.0/30"),
                         AddressAd("object-group ip address NAME")]  # 10.1.1.0/30
            return: [IpNetwork("10.0.0.0 255.255.255.252"), IpNetwork("10.1.1.0 255.255.255.252")]
        """
        ipnets = self.ipnets()
        return [o.with_netmask.replace("/", " ") for o in ipnets]

    def wildcards(self) -> LStr:
        """Wildcards from all addresses in address group.

        :return: Wildcards "A.B.C.D A.B.C.D".
        """
        wildcards: LStr = []
        for addr_o in self._items:
            if addr_o.type == "addrgroup":
                raise TypeError("address group recursion is not supported")
            wildcards_ = addr_o.wildcards()
            wildcards.extend(wildcards_)
        return wildcards

    # =========================== helper =============================

    def _line_to_address(self, line: str) -> OAddressAg:
        """Convert config line to AddressAg object.

        :param line: Address line.
        :return: Address object.

        :example:
            line: "10 host 10.0.0.1"
            return: AddressAg("10 host 10.0.0.1")

        :example: not Address line
            line: "text"
            return: None
        """
        try:
            parsers.parse_address(line)
        except ValueError:
            return None
        addr_o = AddressAg(line=line, platform=self._platform, max_ncwb=self.max_ncwb)
        return addr_o


DAddrGroup = Dict[str, AddrGroup]
LAddrGroup = List[AddrGroup]
UAddrGr = Union[AddressAg, AddrGroup]



"""Functions to create Acl objects From the "show running-config" output."""

import logging
from ipaddress import IPv4Network
from typing import Union

import netports
from vhelpers import vlist

from cisco_acl import helpers as h
from cisco_acl.ace import Ace, LAce
from cisco_acl.ace_group import LUAceg
from cisco_acl.acl import Acl, LAcl
from cisco_acl.addr_group import AddrGroup, LAddrGroup
from cisco_acl.address import Address, LAddress
from cisco_acl.address_ag import AddressAg
from cisco_acl.config_parser import ConfigParser
from cisco_acl.port import Port
from cisco_acl.protocol import Protocol
from cisco_acl.types_ import LDAny, LInt, LStr, DAny, LLStr
from cisco_acl.wildcard import init_max_ncwb

UAddress = Union[Address, AddressAg]


# noinspection PyIncorrectDocstring,DuplicatedCode
def acls(config: str, **kwargs) -> LAcl:
    """Create Acl objects based on the "show running-config" output.

    Support address group objects.
    Each ACE line is treated as an independent Ace (default) or ACE lines can be
    grouped to AceGroup by text in remarks (param `group_by`).

    :param config: Cisco config, "show running-config" output.
    :type config: str

    :param platform: Platform: "asa", "ios", "nxos". Default "ios".
    :type platform: str

    :param version: Software version, default is "0".
    :type version: str

    :param names: Parses only ACLs with specified names, skips any other.
    :type names: List[str]

    :param max_ncwb: Max count of non-contiguous wildcard bits.
    :type max_ncwb: int

    :param indent: ACE lines indentation (default "  ").
    :type indent: str

    :param protocol_nr: Well-known ip protocols as numbers.
        True  - all ip protocols as numbers,
        False - well-known ip protocols as names (default).
    :type protocol_nr: bool

    :param port_nr: Well-known TCP/UDP ports as numbers,
        True  - all tcp/udp ports as numbers.
        False - well-known tcp/udp ports as names (default).
    :type port_nr: bool

    :param group_by: Startswith in remark line. ACEs group, starting from the Remark,
        where line startswith `group_by`, will be applied to the same AceGroup,
        until next Remark that also startswith `group_by`.
    :type group_by: str

    :return: List of Acl objects.
    :rtype: List[Acl]
    """
    platform = h.init_platform(**kwargs)
    version = str(kwargs.get("version") or "")
    group_by = str(kwargs.get("group_by") or "")
    names = kwargs.get("names")
    indent: str = h.init_indent(**kwargs)
    max_ncwb: int = init_max_ncwb(**kwargs)
    protocol_nr = bool(kwargs.get("protocol_nr"))
    port_nr = bool(kwargs.get("port_nr"))

    parser = ConfigParser(config=config, platform=platform, version=version)
    parser.parse_config()
    parsed_acls: LDAny = parser.acls(names=names)

    acl_kwargs = dict(version=version, indent=indent, max_ncwb=max_ncwb,
                      protocol_nr=protocol_nr, port_nr=port_nr)
    acls_: LAcl = [Acl(**acl_kwargs, **d) for d in parsed_acls]  # type: ignore
    _add_addgr_to_aces(acls_, parser)
    if group_by:
        for acl_o in acls_:
            acl_o.group(group_by=group_by)
    return acls_


# noinspection PyIncorrectDocstring,DuplicatedCode
def aces(config: str, **kwargs) -> LUAceg:
    """Create Ace objects based on the "show running-config" output.

    :param config: Cisco config, "show running-config" output.
    :type config: str

    :param platform: Platform: "asa", "ios", "nxos". Default "ios".
    :type platform: str

    :param version: Software version, default is "0".
    :type version: str

    :param max_ncwb: Max count of non-contiguous wildcard bits.
    :type max_ncwb: int

    :param protocol_nr: Well-known ip protocols as numbers.
        True  - all ip protocols as numbers,
        False - well-known ip protocols as names (default).
    :type protocol_nr: bool

    :param port_nr: Well-known TCP/UDP ports as numbers.
        True  - all tcp/udp ports as numbers,
        False - well-known tcp/udp ports as names (default).
    :type port_nr: bool

    :param group_by: Startswith in remark line. ACEs group, starting from the Remark,
        where line startswith `group_by`, will be applied to the same AceGroup,
        until next Remark that also startswith `group_by`.
    :type group_by: str

    :return: List of Ace objects.
    :rtype: List[Ace]
    """
    platform = h.init_platform(**kwargs)
    version = str(kwargs.get("version") or "")
    group_by = str(kwargs.get("group_by") or "")
    max_ncwb: int = init_max_ncwb(**kwargs)
    protocol_nr = bool(kwargs.get("protocol_nr"))
    port_nr = bool(kwargs.get("port_nr"))

    parser = ConfigParser(config=config, platform=platform, version=version)
    parser.parse_config()

    acl_kwargs = dict(version=version, max_ncwb=max_ncwb, protocol_nr=protocol_nr, port_nr=port_nr)
    acl_o = Acl(platform=platform, **acl_kwargs)  # type: ignore
    for line in parser.lines:
        # noinspection PyProtectedMember
        if ace_o := acl_o._line_to_oace(line):
            acl_o.items.append(ace_o)

    if group_by:
        acl_o.group(group_by=group_by)
    return acl_o.items


# noinspection PyIncorrectDocstring
def addrgroups(config: str, **kwargs) -> LAddrGroup:
    """Create AddrGroup objects based on the "show running-config" output.

    :param config: Cisco config, "show running-config" output.
    :type config: str

    :param platform: Platform: "asa", "ios", "nxos". Default "ios".
    :type platform: str

    :param version: Software version, default is "0".
    :type version: str

    :param max_ncwb: Max count of non-contiguous wildcard bits.
    :type max_ncwb: int

    :param indent: Address lines indentation (default "  ").
    :type indent: str

    :return: List of AddrGroup objects.
    :rtype: List[AddrGroup]
    """
    platform = h.init_platform(**kwargs)
    version = str(kwargs.get("version") or "")
    max_ncwb: int = init_max_ncwb(**kwargs)
    indent: str = h.init_indent(**kwargs)

    parser = ConfigParser(config=config, platform=platform, version=version)
    parser.parse_config()

    parsed_addgrs: LDAny = parser.addgrs()
    ag_kwargs = dict(version=version, max_ncwb=max_ncwb, indent=indent)
    addgrs: LAddrGroup = [AddrGroup(**ag_kwargs, **d) for d in parsed_addgrs]  # type: ignore
    return addgrs


# noinspection PyIncorrectDocstring
def range_ports(
        srcports: str = "",
        dstports: str = "",
        line: str = "permit tcp any any",
        platform: str = "",
        port_nr: bool = False,
        port_count: int = 1,
        port_range: bool = True,
        **kwargs,
) -> LStr:
    """Generate ACEs in required range of TCP/UDP source/destination ports.

    :param srcports: Range of TCP/UDP source ports.
    :type srcports: str

    :param dstports: Range of TCP/UDP destination ports.
    :type dstports: str

    :param line: ACE pattern, on whose basis new ACEs will be generated
        (default "permit tcp any any", operator "eq").
    :type line: str

    :param platform: Platform: "asa", "ios", "nxos". Default "ios".
    :type platform: str

    :param port_nr: Well-known TCP/UDP ports as numbers.
        True  - all tcp/udp ports as numbers,
        False - well-known tcp/udp ports as names (default).
    :type port_nr: bool

    :param port_count: Count of ports in ACE lines. Default is 1.
    :type port_count: int

    :param port_range: Transform ACE lines with match-operator "range" to lines with "eq".
        True - Split match-operator "range" and "eq" to different ACE lines, default is True,
        False - Transform all ACEs with "range" to ACEs with "eq" (each port in separate ACE).
    :type port_range: bool

    :return: List of newly generated ACE lines.
    :rtype: List[str]

    :example:
        range_ports("21-23,80") -> ["permit tcp any eq ftp any",
                                    "permit tcp any eq 22 any",
                                    "permit tcp any eq telnet any",
                                    "permit tcp any eq www any"]
    """
    platform = h.init_platform(platform=platform)
    port_count = int(h.init_number(port_count or 1))
    _check_operator_eq_range(line, platform, port_range)

    aces_: LAce = []  # result

    params = {
        "line": line,
        "platform": platform,
        "port_nr": port_nr,
        "port_count": port_count,
        "port_range": port_range,
    }
    for sdst, ports_range in [("src", srcports), ("dst", dstports)]:
        _aces = _range__port(ports_range=ports_range, sdst=sdst, **params)
        aces_.extend(_aces)

    return [o.line for o in aces_]


# noinspection PyIncorrectDocstring
def range_protocols(**kwargs) -> LStr:
    """Generate ACEs in required range of IP protocols.

    :param protocols: Range of IP protocols.
    :type protocols: str

    :param line: ACE pattern, on whose basis new ACEs will be generated.
        (default "permit ip any any")
    :type line: str

    :param platform: Platform: "asa", "ios", "nxos". Default "ios".
    :type platform: str

    :param version: Software version, default is "0".
    :type version: str

    :param protocol_nr: Well-known ip protocols as numbers.
        True  - all ip protocols as numbers,
        False - well-known ip protocols as names (default).
    :type protocol_nr: bool

    :return: List of newly generated ACE lines.
    :rtype: List[str]

    :example:
        range_protocols(protocols="1-2,6", line="permit ip host 10.0.0.1 any") ->
            ["permit icmp host 10.0.0.1 any",
             "permit igmp host 10.0.0.1 any",
             "permit tcp host 10.0.0.1 any"]
    """
    range_ = str(kwargs.get("protocols") or "")
    line = str(kwargs.get("line") or "permit ip any any")
    platform = h.init_platform(**kwargs)
    protocol_nr = bool(kwargs.get("protocol_nr"))

    aces_: LAce = []  # result
    protocols: LInt = netports.iip(range_)
    for proto in protocols:
        ace_o = Ace(line, platform=platform, protocol_nr=protocol_nr)
        ace_o._protocol = Protocol(str(proto), platform=platform, protocol_nr=protocol_nr)
        ace_o.platform = platform
        aces_.append(ace_o)
    return [o.line for o in aces_]


def subnet_of(top: UAddress, bottom: UAddress) -> bool:
    """Check `bottom` address (all ipnets) is subnet of `top` address (any of ipnet).

    :param top: Other address object to check with self address.
    :type top: Union[Address, AddressAg]

    :param bottom: Other address object to check with self address.
    :type bottom: Union[Address, AddressAg]

    :return: True - if address is subnet of `other` address.
    :rtype: bool
    """
    tops_ = top.ipnets()
    bottoms_ = bottom.ipnets()

    for bottom_ in bottoms_:
        for top_ in tops_:
            if bottom_.subnet_of(top_):
                break
        else:
            return False
    return True


# ============================= helper ===============================


def _check_addgr(ace_o, addgrs, address_o, parser) -> bool:
    """Check addresses in address group.

    :return: True - Single Address group present in config.
        False - Address group not found in config or detected multiple groups,
        with the same name.
    """
    ace = ace_o.line
    addrgroup = address_o.addrgroup
    addgrs_ = [o for o in addgrs if o.name == addrgroup]
    count = len(addgrs_)
    if not count:
        line = f"{parser.pattern__object_group()} {addrgroup}"
        msg = f"{ace=} has no addresses, {line=} not found in config"
        logging.warning(msg)
        return False
    if count != 1:
        msg = f"{ace=} has no addresses, found multiple {addrgroup=}, expected 1"
        logging.warning(msg)
        return False
    return True


def _check_operator_eq_range(line: str, platform: str, port_range: bool) -> None:
    """Check if the operator is one of allowed: "eq", "neq", "range".

    :param line: ACE line with interested match-operators to check.

    :param platform: Platform

    :return: None. Raise a ValueError if the operator is invalid.
    """
    ace_o = Ace(line, platform=platform)
    operators = [ace_o.srcport.operator, ace_o.dstport.operator]
    operators = [s for s in operators if s]

    expected = ["eq", "neq", "range"]
    for operator in operators:
        if operator not in expected:
            raise ValueError(f"invalid {operator=}, {expected=}")


def _create_acls_w_acegs(parser: ConfigParser, group_by: str) -> LAcl:
    """Create Acls with AceGroups. Groups ACEs to AceGroup by `group_by` in startswith remarks.

    :param parser: Semi-parsed config.

    :param group_by: Startswith in remark line. ACEs group, starting from the Remark,
        where line startswith `group_by`, will be applied to the same AceGroup,
        until next Remark that also startswith `group_by`.

    :return: List of parsed ACLs.
    """
    parsed_acls = parser.acls()
    acls_w_aceg: LAcl = [Acl(**d) for d in parsed_acls]

    for acl_o in acls_w_aceg:
        acl_o.group(group_by=group_by)

    _add_addgr_to_aces(acls_w_aceg, parser)
    return acls_w_aceg


def _add_addgr_to_aces(acls_: LAcl, parser: ConfigParser) -> None:
    """Add address groups to Ace.srcaddr Ace.dstaddr.

    :param acls_: Side effect.
    """
    parsed_addgrs = parser.addgrs()
    addgrs: LAddrGroup = [AddrGroup(**d) for d in parsed_addgrs]

    for acl_o in acls_:
        _aces: LAce = [o for o in acl_o.items if isinstance(o, Ace)]
        for ace_o in _aces:
            addrs_w_addgr: LAddress = [o for o in (ace_o.srcaddr, ace_o.dstaddr) if o.addrgroup]
            addrs_w_addgr = [o for o in addrs_w_addgr if _check_addgr(ace_o, addgrs, o, parser)]
            for addr_ace_o in addrs_w_addgr:
                addgr_name = addr_ace_o.addrgroup
                addgr_o: AddrGroup = [o for o in addgrs if o.name == addgr_name][0]
                for address_ag_o in addgr_o.items:
                    if not isinstance(address_ag_o, AddressAg):
                        continue
                    address_ag_o.sequence = 0
                    address_ag_d = address_ag_o.data()
                    _convert_ios_addr(address_ag_d)
                    addr_item_o = Address(**address_ag_d)
                    addr_ace_o.items.append(addr_item_o)


def _convert_ios_addr(address_ag_d: DAny) -> None:
    """Convert data of IOS AddressAg to data ready for Address.

    :result: Side effect `address_ag_d`.
    """
    if address_ag_d["platform"] != "ios":
        return
    ipnet = address_ag_d["ipnet"]
    if not isinstance(ipnet, IPv4Network):
        raise TypeError(f"invalid {ipnet=} {IPv4Network} expected")
    wildcard = f"{ipnet.network_address} {ipnet.hostmask}"
    address_ag_d["line"] = wildcard


def _range__port(
        ports_range: str,
        sdst: str,
        line: str,
        platform: str,
        port_nr: bool,
        port_count: int,
        port_range: bool,
) -> LAce:
    """Generate range of TCP/UDP ports with match-operator.

    :param ports_range: Range of TCP/UDP source/destination ports.
    :type ports_range: str

    :param sdst: "src", "dst".
    :type sdst: str

    :param line: ACE line with interested protocol and match operator.
    :type line: str

    :param platform: Platform: "asa", "ios", "nxos". Default "ios".
    :type platform: str

    :param port_nr: Well-known TCP/UDP ports as numbers.
        True  - all tcp/udp ports as numbers,
        False - well-known tcp/udp ports as names (default).
    :type port_nr: bool

    :param port_count: Count of ports in ACE lines. Default is 1.
    :type port_count: int

    :param port_range: Transform ACE lines with match-operator "range" to lines with "eq".
        True - Split match-operator "range" and "eq" to different ACE lines, default is True,
        False - Transform all ACEs with "range" to ACEs with "eq" (each port in separate ACE).
    :type port_range: bool

    :return: List of newly generated Ace objects.
    :rtype: List[Ace]
    """
    aces_: LAce = []  # result

    ports_for_ace: LLStr = _split_range_for_ace(ports_range, port_count, port_range)

    for ports_ in ports_for_ace:
        port = " ".join([f"{i}" for i in ports_])

        ace_o = Ace(line, platform=platform, port_nr=port_nr)
        ace_o.protocol.has_port = True

        # operator
        attr = f"{sdst}port"
        port_o: Port = getattr(ace_o, attr)
        operator = port_o.operator
        if not operator:
            operator = "eq"
            if port.find("-") > -1:
                operator = "range"
                port = port.replace("-", " ")

        port_o = Port(
            line=f"{operator} {port}",
            platform=ace_o.platform,
            protocol=ace_o.protocol.name,
            port_nr=ace_o.port_nr,
        )
        attr = f"_{sdst}port"
        setattr(ace_o, attr, port_o)
        aces_.append(ace_o)

    return aces_


def _split_range_for_ace(ports_range: str, port_count: int, port_range: bool) -> LLStr:
    """Split port_range to strings where each item is for separate ACE.

    :param ports_range: Range of TCP/UDP source/destination ports.
    :type ports_range: str

    :param port_count: Count of ports in ACE lines. Default is 1.
    :type port_count: int

    :param port_range: Transform ACE lines with match-operator "range" to lines with "eq".
        True - Split match-operator "range" and "eq" to different ACE lines, default is True,
        False - Transform all ACEs with "range" to ACEs with "eq" (each port in separate ACE).
    :type port_range: bool

    :return: List of newly generated ports.
    :rtype: List[Ace]
    """
    items: LLStr = []
    ports_i: LStr = []
    ports = ports_range.split(",")
    for port in ports:
        # range
        if not port.isdigit():
            if ports_i:
                items.append(ports_i)
                ports_i = []
            if not port:
                continue
            if port_range:
                items.append([port])
                continue
            ports_ = [str(i) for i in netports.itcp(port)]
            items.append(ports_)
            continue

        # port_count unlimited
        if not port_count:
            ports_i.append(port)
            continue

        # ports
        if len(ports_i) >= port_count:
            items.append(ports_i)
            ports_i = []
        ports_i.append(port)
    else:
        if ports_i:
            items.append(ports_i)

    if not port_range:
        items_ = vlist.flatten(items)
        items = [items_]
        if port_count:
            items = vlist.to_multi(items_, count=port_count)

    return items
